import dask.array as da
import dask.array.core
import dask.array.ma as dama
import xarray
import numpy as np
import math
import re
import argparse
from . import ms_info
from collections import OrderedDict
from shade_ms import log

USE_COUNT_CAT = 0
COUNT_DTYPE = np.int16

def add_options(parser):
    parser.add_argument('--count-cat', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--count-dtype', default='int16', help=argparse.SUPPRESS)

def set_options(options):
    global USE_COUNT_CAT, COUNT_DTYPE
    USE_COUNT_CAT = options.count_cat
    COUNT_DTYPE = getattr(np, options.count_dtype)

def col_to_label(col):
    """Replaces '-' and "/" in column names with palatable characters acceptable in filenames and identifiers"""
    return col.replace("-", "min").replace("/", "div")


class DataMapper(object):
    """This class defines a mapping from a dask group to an array of real values to be plotted"""
    def __init__(self, fullname, unit, mapper, column=None, extras=[], conjugate=False, axis=None, const=False):
        """
        :param fullname:    full name of parameter (real, amplitude, etc.)
        :param unit:        unit string
        :param mapper:      function that maps column data to parameter
        :param column:      fix a column name (e.g. TIME, UVW. Not for visibility columns.)
        :param extras:      extra arguments needed by mapper (e.g. ["freqs", "wavel"])
        :param conjugate:   sets conjugation flag
        :param axis:        which axis the parameter represets (0 time, 1 freq), if 1-dimensional
        :param const:       if True, axis is constant (does not depend on MS rows)
        """
        assert mapper is not None
        self.fullname, self.unit, self.mapper, self.column, self.extras = fullname, unit, mapper, column, extras
        self.conjugate = conjugate
        self.axis = axis
        # if
        self.const = const


_identity = lambda x:x

# this dict maps short axis names into full DataMapper objects
data_mappers = OrderedDict(
    _     = DataMapper("", "", _identity),
    amp   = DataMapper("amplitude", "", abs),
    logamp = DataMapper("log-amplitude", "", lambda x:da.log10(abs(x))),
    phase = DataMapper("phase", "deg", lambda x:da.arctan2(da.imag(x), da.real(x))*180/math.pi),
    real  = DataMapper("real", "", da.real),
    imag  = DataMapper("imag", "", da.imag),
    TIME  = DataMapper("time", "s", axis=0, column="TIME", mapper=_identity),
    ROW   = DataMapper("row number", "", column=False, axis=0, extras=["rows"], mapper=lambda x,rows: rows),
    BASELINE = DataMapper("baseline", "", column=False, axis=0, extras=["baselines"], mapper=lambda x,baselines: baselines),
    BASELINE_M = DataMapper("baseline", "", column=False, axis=0, extras=["baselines"], mapper=lambda x,baselines: baselines),
    CORR  = DataMapper("correlation", "", column=False, axis=0, extras=["corr"], mapper=lambda x,corr: corr, const=True),
    CHAN  = DataMapper("channel", "", column=False, axis=1, extras=["chans"], mapper=lambda x,chans: chans, const=True),
    FREQ  = DataMapper("frequency", "Hz", column=False, axis=1, extras=["freqs"], mapper=lambda x, freqs: freqs, const=True),
    WAVEL = DataMapper("wavelength", "m", column=False, axis=1, extras=["wavel"], mapper=lambda x, wavel: wavel, const=True),
    UV    = DataMapper("uv-distance", "wavelengths", column="UVW", extras=["wavel"],
                  mapper=lambda uvw, wavel: da.sqrt((uvw[:,:2]**2).sum(axis=1))/wavel),
    U     = DataMapper("u", "wavelengths", column="UVW", extras=["wavel"],
                  mapper=lambda uvw, wavel: uvw[:, 0] / wavel,
                 conjugate=True),
    V     = DataMapper("v", "wavelengths", column="UVW", extras=["wavel"],
                 mapper=lambda uvw, wavel: uvw[:, 1] / wavel,
                 conjugate=True),
    W     = DataMapper("w", "wavelengths", column="UVW", extras=["wavel"],
                 mapper=lambda uvw, wavel: uvw[:, 2] / wavel,
                 conjugate=True),
)


class DataAxis(object):
    """
    Represents a data axis that can be plotted: a combination of column, function and correlation
    """
    # dict of all registered axes
    all_axes = OrderedDict()
    # set of labels (used to make unique labels)
    all_labels = set()

    @classmethod
    def is_legit_colspec(cls, colspec, ms):
        match = re.fullmatch(r"(\w+)([*/+-])(\w+)", colspec)
        if match:
            return match.group(1) in ms.valid_columns and match.group(3) in ms.valid_columns
        else:
            return colspec in ms.valid_columns

    @classmethod
    def parse_datum_spec(cls, axis_spec, default_column=None, ms=None):
        # figure out the axis specification
        function = column = corr = None
        specs = axis_spec.split(":", 2)
        for spec in specs:
            if spec is None:
                raise ValueError(f"invalid axis specification '{axis_spec}'")
            if spec in data_mappers:
                if function is not None:
                    raise ValueError(f"function specified twice in '{axis_spec}'")
                function = spec
            elif DataAxis.is_legit_colspec(spec, ms):
                if column is not None:
                    raise ValueError(f"column specified twice in '{axis_spec}'")
                column = spec
            else:
                if re.fullmatch(r"\d+", spec):
                    corr1 = int(spec)
                else:
                    corr1 = ms.all_corr[spec]
                    if corr1 is None:
                        raise ValueError(f"invalid axis specification '{axis_spec}'")
                if corr is not None:
                    raise ValueError(f"correlation specified twice in '{axis_spec}'")
                corr = corr1
        # defaults?
        if function is None:
            if column or corr:
                function = "_"
            else:
                raise ValueError(f"invalid axis specification '{axis_spec}'")
        # if mapper has a "False" column, it's a fake column...
        mapper_column = data_mappers[function].column
        column = (column or mapper_column or default_column) if mapper_column is not False else None
        has_corr_axis = (column and (column.endswith("DATA") or column.endswith("SPECTRUM"))) or function == "CORR"
        if not has_corr_axis:
            if corr is not None:
                raise ValueError(f"'{axis_spec}': can't specify correlation when column is '{column}'")
            corr = False  # corr=False marks a datum without a correlation axis
        return function, column, corr, (has_corr_axis and corr is None)

    @classmethod
    def register(cls, function, column, corr, ms, minmax=None, ncol=None, subset=None, minmax_cache=None):
        """
        Registers a data axis, which ultimately ends up as a column in the assembled dataframe.
        For multiple plots, we want to reuse the same information (assuming the same
        clipping limits, etc.) so we have a dictionary of axis definitions here.

        Columns selects a column to operate on.
        Function selects a mapping (see datamappers below).
        Corr selects a correlation (or a Stokes product such as I, Q,...)
        minmax sets axis clipping levels
        ncol discretizes the axis into N colours between min and max
        minmax_cache provides a dict of cached min/max values, which will be looked up via the label, if minmax
                     is not explicitly set
        """
        # form up label
        label  = "{}_{}_{}".format(col_to_label(column or ''), function, corr)
        minmax = tuple(minmax) if minmax is not None else (None, None)
        key = label, minmax, ncol
        # see if this axis definition already exists, else create new one
        if key in cls.all_axes:
            return cls.all_axes[key]
        else:
            # see if minmax should be loaded
            if (minmax is None or tuple(minmax) == (None,None)) and minmax_cache and label in minmax_cache:
                log.info(f"loading {label} min/max from cache")
                minmax = minmax_cache[label]

            label0, i = label, 0
            while label in cls.all_labels:
                i += 1
                label = f"{label}_{i}"
            cls.all_labels.add(label)
            axis = cls.all_axes[key] = DataAxis(column, function, corr, ms, minmax, ncol, label, subset=subset)
            return axis

    def __init__(self, column, function, corr, ms, minmax=None, ncol=None, label=None, subset=None):
        """See register() class method above. Not called directly."""
        self.name = ":".join([str(x) for x in (function, column, corr, minmax, ncol) if x is not None ])
        self.ms = ms
        self.function = function        # function to apply to column (see list of DataMappers below)
        self.corr     = corr if corr != "all" else None
        self.nlevels  = ncol
        self.minmax   = tuple(minmax) if minmax is not None else (None, None)
        self._minmax_autorange = (self.minmax == (None, None))

        self.label    = label
        self._corr_reduce = None
        self._is_discrete = None
        self.mapper = data_mappers[function]

        # if set, axis is discrete and labelled
        self.discretized_labels = None

        # for discrete axes: if a subset of N indices is explicitly selected for plotting, then this
        # is a list of the selected indices, of length N
        self.subset_indices = None
        # ...and this is a dask array that maps selected indices into bins 0...N-1, and all other values into bin N
        self.subset_remapper = None
        # ...and this is the maximum valid index in MS
        maxind = None

        # columns with labels?
        if function == 'CORR' or function == 'STOKES':
            corrset = set(subset.corr.names)
            corrset1 = corrset - set("IQUV")
            if corrset1 == corrset:
                name = "Correlation"
            elif not corrset1:
                name = "Stokes"
            else:
                name = "Correlation or Stokes"
            # special case of "corr" if corr is fixed: return constant value fixed here
            # When corr is fixed, we're creating a mapper for a know correlation. When corr is not fixed,
            # we're creating one for a mapper that will iterate over correlations
            if corr is not None:
                self.mapper = DataMapper(name, "", column=False, axis=-1, mapper=lambda x: corr)
            self.subset_indices = subset.corr
            maxind = ms.all_corr.numbers[-1]
        elif column == "FIELD_ID":
            self.subset_indices = subset.field
            maxind = ms.field.numbers[-1]
        elif column == "ANTENNA1" or column == "ANTENNA2":
            self.subset_indices = subset.ant
            maxind = ms.antenna.numbers[-1]
        elif column == "SCAN_NUMBER":
            self.subset_indices = subset.scan
            maxind = ms.scan.numbers[-1]
        elif function == "BASELINE":
            self.subset_indices = subset.baseline
            maxind = ms.baseline.numbers[-1]
        elif function == "BASELINE_M":
            bl_subset = set(subset.baseline.numbers)   # active baselines
            numbers = [i for i in ms.baseline_m.numbers if i in bl_subset]
            names = [bl for i, bl in zip(ms.baseline_m.numbers, ms.baseline_m.names) if i in bl_subset]
            self.subset_indices = ms_info.NamedList("baseline_m", names, numbers)
            maxind = ms.baseline.numbers[-1]
        elif column == "FLAG" or column == "FLAG_ROW":
            self.discretized_labels = ["F", "T"]

        # make a remapper
        if self.subset_indices is not None:
            # if the mapping from indices to bins 1:1?
            subind = np.array(self.subset_indices.numbers)
            identity = subind[0] == 0 and ((subind[1:]-subind[:-1]) == 1).all()
            # If mapping is not 1:1, or subset is short of full set, then we need a remapper.
            # Map indices in subset into their ordinal numbers in the subset (0...N-1), and all other indices to N
            if len(self.subset_indices) < maxind+1 or not identity:
                remapper = np.full(maxind+1, len(self.subset_indices))
                for i, index in enumerate(self.subset_indices.numbers):
                    remapper[index] = i
                self.subset_remapper = da.array(remapper)
            self.discretized_labels = self.subset_indices.names
            self.subset_indices = self.subset_indices.numbers

        if self.discretized_labels:
            self._is_discrete = True

        # axis name
        self.fullname = self.mapper.fullname or column or ''

        # if labels were set up, adjust nlevels (but only down, never up)
        if self.discretized_labels is not None and self.nlevels is not None:
            self.nlevels = min(len(self.discretized_labels), self.nlevels)

        if self.function == "_":
            self.function = ""
        self.conjugate = self.mapper.conjugate
        self.timefreq_axis = self.mapper.axis

        # setup columns
        self._ufunc = None
        self.columns = ()

        # does the mapper have no column (i.e. frequency)?
        if self.mapper.column is False:
            log.info(f'axis: {function}, range {self.minmax}, discretization {self.nlevels}')
        # does the mapper have a fixed column? This better be consistent
        elif self.mapper.column is not None:
            # if a mapper (such as "uv") implies a fixed column name, make sure it's consistent with what the user said
            if column and self.mapper.column != column:
                raise ValueError(f"'{function}' not applicable with column {column}")
            self.columns = (column,)
            log.info(f'axis: {function}({column}), range {self.minmax}, discretization {self.nlevels}')
        # else arbitrary column
        else:
            log.info(f'axis: {function}({column}), corr {self.corr}, range {self.minmax}, discretization {self.nlevels}')
            # check for column arithmetic
            match = re.fullmatch(r"(\w+)([*/+-])(\w+)", column)
            if match:
                self.columns = (match.group(1), match.group(3))
                # look up dask ufunc corresponding to arithmetic op
                self._ufunc = {'+': da.add, '*': da.multiply, '-': da.subtract, '/': da.divide}[match.group(2)]
            else:
                self.columns = (column,)

    @property
    def is_discrete(self):
        return self._is_discrete

    def get_column_data(self, group):
        """Given a dask group, returns dask array corresponding to column setting"""
        if not self.columns:
            return None
        try:
            data = getattr(group, self.columns[0])
            if self._ufunc:
                return self._ufunc(data, getattr(group, self.columns[1]))
            return data
        except AttributeError:
            raise NameError("column {} not found in group".format(" or ".join(self.columns)))

    def get_value(self, group, corr, extras, flag, flag_row, chanslice):
        coldata = self.get_column_data(group)
        # correlation may be pre-set by plot type, or may be passed to us
        corr = self.corr if self.corr is not None else corr
        # apply correlation reduction
        if coldata is not None and coldata.ndim == 3:
            assert corr is not None
            # the mapper can't have a specific axis set
            if self.mapper.axis is not None:
                raise TypeError(f"{self.name}: unexpected column with ndim=3")
            coldata = self.ms.corr_data_mappers[corr](coldata)
        # apply mapping function
        mapper = self.mapper
        # complex values with an identity mapper get an amp mapper assigned to them by default
        if np.iscomplexobj(coldata) and mapper is data_mappers["_"]:
            mapper = data_mappers["amp"]
        coldata = mapper.mapper(coldata, **{name:extras[name] for name in self.mapper.extras })
        # for a constant axis, compute minmax on the fly
        if mapper.const and self._minmax_autorange:
            if np.isscalar(coldata):
                min1 = max1 = coldata
            else:
                min1, max1 = coldata.data.min(), coldata.data.max()
            self.minmax = min(self.minmax[0], min1) if self.minmax[0] is not None else min1, \
                          min(self.minmax[1], max1) if self.minmax[1] is not None else max1
        # scalar is just a scalar
        if np.isscalar(coldata):
            coldata = da.array(coldata)
            flag = None
        else:
            # apply channel slicing, if there's a channel axis in the array (and the array is a DataArray)
            if type(coldata) is xarray.DataArray and 'chan' in coldata.dims:
                coldata = coldata[dict(chan=chanslice)]
            # determine flags -- start with original flags
            if flag is not None:
                if coldata.ndim == 2:
                    flag = self.ms.corr_flag_mappers[corr](flag)
                elif coldata.ndim == 1:
                    if not self.mapper.axis:
                        flag = flag_row
                    elif self.mapper.axis == 1:
                        flag = None
                # shapes must now match
                if flag is not None and coldata.shape != flag.shape:
                    raise TypeError(f"{self.name}: unexpected column shape")
        # # discretize
        # if self.nlevels:

        if coldata.dtype is bool or np.issubdtype(coldata.dtype, np.integer):
            if self._is_discrete is False:
                raise TypeError(f"{self.label}: column changed from continuous-valued to discrete. This is a bug, or a very weird MS.")
            self._is_discrete = True
            # do we need to apply a remapping?
            if self.subset_remapper is not None:
                if type(coldata) is not dask.array.core.Array:  # could be xarray backed by dask array
                    coldata = coldata.data
                coldata = self.subset_remapper[coldata]
                bad_bins = da.greater_equal(coldata, len(self.subset_indices))
                if flag is None:
                    flag = bad_bins
                else:
                    flag = da.logical_or(flag.data, bad_bins)
        else:
            if self._is_discrete is True:
                raise TypeError(f"{self.label}: column chnaged from discrete to continuous-valued. This is a bug, or a very weird MS.")
            self._is_discrete = False

        # Ensure dask arrays for creating dask masked arrays
        if isinstance(coldata, xarray.DataArray):
            coldata = coldata.data

        if isinstance(flag, xarray.DataArray):
            flag = flag.data

        bad_data = da.logical_not(da.isfinite(coldata))
        if flag is not None:
            return dama.masked_array(coldata, da.logical_or(flag, bad_data))
        else:
            return dama.masked_array(coldata, bad_data)




# -*- coding: future_fstrings -*-

import matplotlib
matplotlib.use('agg')

import dask.diagnostics
import datetime
import json
import logging
import os
import re
import shade_ms
import sys
import dask.diagnostics
from contextlib import contextmanager
import json
import yaml
import time
import warnings

from contextlib import contextmanager

from . import DEFAULT_CNUM, DEFAULT_NUM_RENDERS
from . import cli, parse_plot_spec, parse_slice_spec
from . import data_plots, data_mappers

from .data_mappers import DataAxis, col_to_label
from .ms_info import MSInfo

from shade_ms import log, separator


# empty context manager for use when profiling is off
@contextmanager
def nullcontext(enter_result=None):
    yield enter_result


def main(argv):

    # TODO: I want to inspect a different time profile method
    clock_start = time.time()

    # default # of CPUs

    # ---  dealing with input arguments ---
    parser, optimization_opts = cli()
    # various hidden performance-testing options
    data_mappers.add_options(optimization_opts)
    data_plots.add_options(optimization_opts)
    options = parser.parse_args(argv)
    if options.debug:
        shade_ms.log_console_handler.setLevel(logging.DEBUG)

    # pass options to shade_ms
    data_mappers.set_options(options)
    data_plots.set_options(options)

    # set colormaps
    cmap = data_plots.get_colormap(options.cmap)
    bmap = data_plots.get_colormap(options.bmap)
    dmap = data_plots.get_colormap(options.dmap)

    # figure our list of plots to make
    # define plot parameters
    [xaxes, yaxes, columns, caxes, aaxes, areds,
     xmins, xmaxs, ymins, ymaxs, amins, amaxs, cmins, cmaxs, cnums] = parse_plot_spec(parser,
                                                                                      options)


    # check markup arguments
    extra_markup = []
    for funcname, funcargs in options.markup or []:
        from matplotlib.axes import Axes
        if funcname not in dir(Axes):
            parser.error(f"unknown function given in --markup {funcname}")
        args = yaml.safe_load(funcargs)
        kwargs = []
        if type(args) not in {list, dict}:
            parser.error(f"invalid arguments to --markup {funcname} {funcargs}")
        if type(args) is list:
            if len(args) > 0 and type(args[-1]) is dict:
                kwargs = args[-1]
                args = args[:-1]
        else:
            kwargs = args
            args = []
        extra_markup.append((funcname, args, kwargs))

    # check vline and hline
    for attr in 'hline', 'vline':
        if getattr(options, attr, None):
            for spec in getattr(options, attr).split(","):
                match = re.match('(.*?)((--|-|-\\.|:)([a-zA-Z#].*)?)?$', spec)
                try:
                    coord = float(match.group(1))
                except ValueError:
                    coord = None
                    parser.error(f"invalid --{attr} setting '{spec}'")
                linestyle = match.group(3) or '-'
                extra_markup.append((f"ax{attr}", [coord], dict(ls=linestyle,
                                                                color=match.group(4) or "black")))

    # re-check that kwargs are valid
    for funcname, args, kwargs in extra_markup:
        log.info(f"markup: {funcname} *{args} **{kwargs})")
        if not all([re.match('^\w+$', kw) for kw in kwargs.keys()]):
            log.error("the above is not a valid markup specification, please fix")
            sys.exit(1)

    # check chan slice
    try:
        chanslice, chanslice_spec = parse_slice_spec(options.chan)
    except ValueError:
        # parser.error(f"invalid selection --{'chan'} {options.chan}")
        parser.error(f"invalid selection --chan {options.chan}")

    # issue warning if only a single antenna is specified
    num_ants_warning = False
    if options.ant_num:
        specs = options.ant_num.split(',')
        if len(specs) == 1 and re.fullmatch(r"\d+", specs[0]):
            ant_spec = "ant{}".format(int(specs[0])-1)
            num_ants_warning = True
        # check ant selection before processing
        for spec in options.ant_num.split(","):
            try:
                antslice, antslice_spec = parse_slice_spec(spec)
            except ValueError:
                parser.error(f"invalid selection --{'ant-num'} {options.ant_num}")
    elif not 'all' in options.ant:
        specs = options.ant.split(',')
        if len(specs) == 1:
            ant_spec = "{}".format(specs[0])
            num_ants_warning = True
    if num_ants_warning:
        msg = f"Suggested usage: '--baseline {ant_spec}-*' to select all baselines to {ant_spec}."
        warnings.warn(msg, category=UserWarning)

    log.info(" ".join(sys.argv))

    separator()
    # ---  dealing with input arguments ---

    ms = MSInfo(options.ms, log=log)

    separator()
    log.info(": Data selected for plotting:")

    group_cols = ['FIELD_ID', 'DATA_DESC_ID']

    # --- building SQL query --
    mytaql = []

    class Subset(object):
        pass
    subset = Subset()

    if not 'all' in options.ant or options.ant_num:
        if options.ant_num:
            ant_subset = set()
            for spec in options.ant_num.split(","):
                if re.fullmatch(r"\d+", spec):
                    ant_subset.add(int(spec))
                else:
                    antslice, antslice_spec = parse_slice_spec(spec)
                    ant_subset.update(ms.all_antenna.numbers[antslice])
            subset.ant = ms.antenna.get_subset(sorted(ant_subset))
        else:
            subset.ant = ms.antenna.get_subset(options.ant)
        log.info(f"Antenna name(s)  : {' '.join(subset.ant.names)}")
        antnum_set = f"[{','.join(map(str, subset.ant.numbers))}]"
        mytaql.append(f"ANTENNA1 IN {antnum_set} && ANTENNA2 IN {antnum_set}")
    else:
        subset.ant = ms.antenna
        log.info('Antenna(s)       : all')

    if options.baseline == 'all':
        log.info('Baseline(s)      : all')
        subset.baseline = ms.baseline
    elif options.baseline == 'noautocorr':
        log.info('Baseline(s)      : all except autocorrelations')
        subset.baseline = ms.all_baseline.get_subset([i for i in ms.baseline.numbers if ms.baseline_lengths[i] !=0 ])
        mytaql.append("ANTENNA1!=ANTENNA2")
    elif options.baseline == 'autocorr':
        log.info('Baseline(s)      : autocorrelations')
        subset.baseline = ms.all_baseline.get_subset([i for i in ms.baseline.numbers if ms.baseline_lengths[i] == 0])
        mytaql.append("ANTENNA1==ANTENNA2")
    else:
        bls = set()
        a1a2 = set()
        for blspec in options.baseline.split(","):
            match = re.fullmatch(r"(\w+)-(\w*|[*])", blspec)
            ant1 = match and ms.antenna[match.group(1)]
            ant2 = match and (ms.antenna[match.group(2)] if match.group(2) not in ['', '*'] else '*')
            if ant1 is None or ant2 is None:
                raise ValueError(f"invalid baseline '{blspec}'")
            if ant2 == '*':
                ant2set = ms.all_antenna.numbers
            else:
                ant2set = [ant2]
            # loop
            for ant2 in ant2set:
                a1, a2 = min(ant1, ant2), max(ant1, ant2)
                a1a2.add((a1, a2))
                bls.add(ms.baseline_number(a1, a2))
        # group_cols.append('ANTENNA1')
        subset.baseline = ms.all_baseline.get_subset(sorted(bls))
        log.info(f"Baseline(s)      : {' '.join(subset.baseline.names)}")
        mytaql.append("||".join([f'(ANTENNA1=={ant1}&&ANTENNA2=={ant2})||(ANTENNA1=={ant2}&&ANTENNA2=={ant1})'
                                 for ant1, ant2 in a1a2]))

    if options.field != 'all':
        subset.field = ms.field.get_subset(options.field)
        log.info(f"Field(s)         : {' '.join(subset.field.names)}")
        # here for now, workaround for https://github.com/ska-sa/dask-ms/issues/100, should be inside if clause
        mytaql.append("||".join([f'FIELD_ID=={fld}' for fld in subset.field.numbers]))
    else:
        subset.field = ms.field
        log.info('Field(s)         : all')

    if options.spw != 'all':
        subset.spw = ms.spw.get_subset(options.spw)
        log.info(f"SPW(s)           : {' '.join(subset.spw.names)}")
        mytaql.append("||".join([f'DATA_DESC_ID=={ddid}' for ddid in subset.spw.numbers]))
    else:
        subset.spw = ms.spw
        log.info(f'SPW(s)           : all')

    if options.scan != 'all':
        subset.scan = ms.scan.get_subset(options.scan, allow_numeric_indices=False)
        log.info(f"Scan(s)          : {' '.join(subset.scan.names)}")
        mytaql.append("||".join([f'SCAN_NUMBER=={n}' for n in subset.scan.numbers]))
    else:
        subset.scan = ms.scan
        log.info('Scan(s)          : all')
    if options.iter_scan:
        group_cols.append('SCAN_NUMBER')

    if chanslice == slice(None):
        log.info('Channels         : all')
    else:
        log.info(f"Channels         : {':'.join(str(x) if x is not None else '' for x in chanslice_spec)}")

    mytaql = ' && '.join([f"({t})" for t in mytaql]) if mytaql else ''
  # --- building SQL query --

    if options.corr == "ALL":
        subset.corr = ms.corr
    else:
        # recognize shortcut when it's just Stokes indices, convert to list
        if re.fullmatch(r"[IQUV]+", options.corr):
            options.corr = ",".join(options.corr)
        subset.corr = ms.all_corr.get_subset(options.corr)
    log.info(f"Corr/Stokes      : {' '.join(subset.corr.names)}")

    separator()

    # check minmax cache
    msbase = os.path.splitext(os.path.basename(options.ms))[0]
    cache_file = options.lim_file.format(ms=msbase)
    if options.dir and not "/" in cache_file:
        cache_file = os.path.join(options.dir, cache_file)

    # try to load the minmax cache file
    if not os.path.exists(cache_file):
        minmax_cache = {}
    else:
        log.info(f"loading minmax cache from {cache_file}")
        try:
            minmax_cache = json.load(open(cache_file, "rt"))
            if type(minmax_cache) is not dict:
                raise TypeError("cache content is not a dict")
        except Exception as exc:
            log.error(f"error reading cache file: {exc}. Minmax cache will be reset.")
            minmax_cache = {}

    # figure out list of plots to make
    all_plots = []

    # This will be True if any of the specified axes change with correlation
    have_corr_dependence = False

    # now go create definitions
    for xaxis, yaxis, default_column, caxis, aaxis, ared, xmin, xmax, ymin, ymax, amin, amax, cmin, cmax, cnum in \
        zip(xaxes, yaxes, columns, caxes, aaxes, areds, xmins, xmaxs, ymins, ymaxs, amins, amaxs, cmins, cmaxs, cnums):
        # get axis specs
        xspecs = DataAxis.parse_datum_spec(xaxis, default_column, ms=ms)
        yspecs = DataAxis.parse_datum_spec(yaxis, default_column, ms=ms)
        aspecs = DataAxis.parse_datum_spec(aaxis, default_column, ms=ms) if aaxis else [None] * 4
        cspecs = DataAxis.parse_datum_spec(caxis, default_column, ms=ms) if caxis else [None] * 4
        # parse axis specifications
        xfunction, xcolumn, xcorr, xitercorr = xspecs
        yfunction, ycolumn, ycorr, yitercorr = yspecs
        afunction, acolumn, acorr, aitercorr = aspecs
        cfunction, ccolumn, ccorr, citercorr = cspecs
        # does anything here depend on correlation?
        datum_itercorr = (xitercorr or yitercorr or aitercorr or citercorr)
        if datum_itercorr:
            have_corr_dependence = True
        if "FLAG" in (xcolumn, ycolumn, acolumn, ccolumn) or "FLAG_ROW" in (xcolumn, ycolumn, acolumn, ccolumn):
            if not options.noflags:
                log.info(": plotting a flag column implies that flagged data will not be masked")
                options.noflags = True

        # do we iterate over correlations/Stokes to make separate plots now?
        if datum_itercorr and options.iter_corr:
            corr_list = subset.corr.numbers
        else:
            corr_list = [None]

        def describe_corr(corrvalue):
            """Returns list of correlation labels corresponding to this corr setting"""
            if corrvalue is None:
                return subset.corr.names
            elif corrvalue is False:
                return []
            else:
                return [ms.all_corr.names[corrvalue]]

        for corr in corr_list:
            plot_xcorr = corr if xcorr is None else xcorr  # False if no corr in datum, None if all, else set to iterant or to fixed value
            plot_ycorr = corr if ycorr is None else ycorr
            plot_acorr = corr if acorr is None else acorr
            plot_ccorr = corr if ccorr is None else ccorr
            xdatum = DataAxis.register(xfunction, xcolumn, plot_xcorr, ms=ms, minmax=(xmin, xmax), subset=subset,
                                       minmax_cache=minmax_cache if options.xlim_load else None)
            ydatum = DataAxis.register(yfunction, ycolumn, plot_ycorr, ms=ms, minmax=(ymin, ymax), subset=subset,
                                       minmax_cache=minmax_cache if options.ylim_load else None)
            adatum = afunction and DataAxis.register(afunction, acolumn, plot_acorr, ms=ms,
                                                     minmax=(amin, amax), subset=subset)
            cdatum = cfunction and DataAxis.register(cfunction, ccolumn, plot_ccorr, ms=ms,
                                                     minmax=(cmin, cmax), ncol=cnum, subset=subset,
                                                     minmax_cache=minmax_cache if options.clim_load else None)

            # figure out plot properties -- basically construct a descriptive name and label
            # looks complicated, but we're just trying to figure out what to put in the plot title...
            props = dict()
            titles = []
            labels = []
            # start with column and correlation(s)
            if ycolumn and not ydatum.mapper.column:   # only put column if not fixed by mapper
                titles.append(ycolumn)
                labels.append(col_to_label(ycolumn))
            titles += describe_corr(plot_ycorr)
            labels += describe_corr(plot_ycorr)
            if ydatum.mapper.fullname:
                titles += [ydatum.mapper.fullname]
            titles += ["vs"]
            if ydatum.function:
                labels.append(ydatum.function)
            # add x column/subset.corr, if different
            if xcolumn and (xcolumn != ycolumn or not xdatum.function) and not xdatum.mapper.column:
                titles.append(xcolumn)
                labels.append(col_to_label(xcolumn))
            if plot_xcorr is not plot_ycorr:
                titles += describe_corr(plot_xcorr)
                labels += describe_corr(plot_xcorr)
            if xdatum.mapper.fullname:
                titles += [xdatum.mapper.fullname]
            if xdatum.function:
                labels.append(xdatum.function)
            props['title'] = " ".join(titles)
            props['label'] = "-".join(labels)
            # build up intensity label
            if afunction:
                titles, labels = [ared], [ared]
                if acolumn and (acolumn != xcolumn or acolumn != ycolumn) and adatum.mapper.column is None:
                    titles.append(acolumn)
                    labels.append(col_to_label(acolumn))
                if plot_acorr is not None and plot_acorr is not False and \
                        (plot_acorr is not plot_xcorr or plot_acorr is not plot_ycorr):
                    titles += describe_corr(plot_acorr)
                    labels += describe_corr(plot_acorr)
                titles += [adatum.mapper.fullname]
                if adatum.function:
                    labels.append(adatum.function)
                props['alpha_title'] = " ".join(titles)
                props['alpha_label'] = "-".join(labels)
            else:
                props['alpha_title'] = props['alpha_label'] = ''
            # build up color-by label
            if cfunction:
                titles, labels = [], []
                if ccolumn and (ccolumn != xcolumn or ccolumn != ycolumn) and cdatum.mapper.column is None:
                    titles.append(ccolumn)
                    labels.append(col_to_label(ccolumn))
                if plot_ccorr is not None and plot_ccorr is not False and \
                        (plot_ccorr is not plot_xcorr or plot_ccorr is not plot_ycorr):
                    titles += describe_corr(plot_ccorr)
                    labels += describe_corr(plot_ccorr)
                if cdatum.mapper.fullname:
                    titles.append(cdatum.mapper.fullname)
                if cdatum.function:
                    labels.append(cdatum.function)
                props['color_title'] = " ".join(titles)
                props['color_label'] = "-".join(labels)
                props['color_modulo'] = f"(into {cdatum.nlevels} colours)"
            else:
                props['color_title'] = props['color_label'] = ''

            all_plots.append((props, xdatum, ydatum, adatum, ared, cdatum))
            log.debug(f"adding plot for {props['title']}")

    # reset minmax cache if requested
    if options.lim_save_reset:
        minmax_cache = {}

    join_corrs = not options.iter_corr and len(subset.corr) > 1 and have_corr_dependence

    log.info('                 : you have asked for {} plots employing {} unique datums'.format(len(all_plots),
                                                                                                len(DataAxis.all_axes)))
    if not len(all_plots):
        sys.exit(0)

    log.debug(f"taql is {mytaql}, group_cols is {group_cols}, join subset.corr is {join_corrs}")

    dataframes, index_subsets, np = \
        data_plots.get_plot_data(ms, group_cols, mytaql, ms.chan_freqs,
                                 chanslice=chanslice, subset=subset,
                                 noflags=options.noflags, noconj=options.noconj,
                                 iter_field=options.iter_field, iter_spw=options.iter_spw,
                                 iter_scan=options.iter_scan, iter_ant=options.iter_ant,
                                 iter_baseline=options.iter_baseline,
                                 join_corrs=join_corrs,
                                 row_chunk_size=options.row_chunk_size)
    if len(dataframes) < 1:
        log.warn("No data for selection subset")
    log.info(f": rendering {len(dataframes)} dataframes with {np:.3g} points into {len(all_plots)} plot types")

    ## each dataframe is an instance of the axes being iterated over -- on top of that, we need to iterate over plot types

    # dictionary of substitutions for filename and title
    keys = {}
    keys['ms'] = os.path.basename(os.path.splitext(options.ms.rstrip("/"))[0])
    keys['timestamp'] = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    # dictionary of titles for these identifiers
    titles = dict(field="", field_num="field", scan="scan", spw="spw", antenna="ant", colortitle="coloured by")

    keys['field_num'] = subset.field.numbers if options.field != 'all' else ''
    keys['field'] = subset.field.names if options.field != 'all' else ''
    keys['scan'] = subset.scan.names if options.scan != 'all' else ''
    keys['ant'] = subset.ant.names if options.ant != 'all' else ''  ## TODO: also handle ant-num settings
    keys['spw'] = subset.spw.names if options.spw != 'all' else ''
    keys['baseline'] = None

    keys['suffix'] = suffix = options.suffix.format(**options.__dict__) if options.suffix else ''
    keys['_suffix'] = f".{suffix}" if suffix else ''

    def generate_string_from_keys(template, keys, listsep=" ", titlesep=" ", prefix=" "):
        """Converts list of keys into a string suitable for plot titles or filenames.
        listsep is used to separate list elements: " " for plot titles, "_" for filenames.
        titlesep is used to separate names from values: " " for plot titles, "_" for filenames
        prefix is used to prefix {_key} substitutions: ", " for titles, "-" for filenames
        """
        full_keys = keys.copy()
        # convert lists to strings, add Capitalized keys with titles
        for key, value in keys.items():
            capkey = key.title()
            if value == '':
                full_keys[capkey] = ''
            else:
                if type(value) is list:                               # e.g. scan=[1,2] becomes scan="1 2"
                    full_keys[key] = value = listsep.join(map(str, value))
                if key in titles:                                     # e.g. Scan="scan 1 2"
                    full_keys[capkey] = f"{titles[key]}{titlesep}{value}"
                else:
                    full_keys[capkey] = value
        # add _keys which supply prefixes
        full_keys.update({f"_{key}": (f"{prefix}{value}" if value else '') for key, value in full_keys.items()})
        # finally, format
        return template.format(**full_keys)

    jobs = []
    executor = None

    def render_single_plot(df, subset, xdatum, ydatum, adatum, ared, cdatum, pngname, title, xlabel, ylabel):
        """Renders a single plot. Make this a function since we might call it in parallel"""
        log.info(f": rendering {pngname}")
        normalize = options.norm
        if normalize == "auto":
            normalize = "log" if cdatum is not None else ("eq_hist" if adatum is None else 'linear')
        if options.profile:
            context = dask.diagnostics.ResourceProfiler
        else:
            context = nullcontext
        with context() as profiler:
            result = data_plots.create_plot(df, subset, xdatum, ydatum, adatum, ared, cdatum,
                                      cmap=cmap, bmap=bmap, dmap=dmap, normalize=normalize,
                                      min_alpha=options.min_alpha,
                                      saturate_alpha=options.saturate_alpha,
                                      saturate_percentile=options.saturate_perc,
                                      xlabel=xlabel, ylabel=ylabel, title=title, pngname=pngname,
                                      extra_markup=extra_markup,
                                      minmax_cache=minmax_cache,
                                      options=options)
        if result:
            log.info(f'                 : wrote {pngname}')
            if profiler is not None:
                profile_file = os.path.splitext(pngname)[0] + ".prof.html"
                dask.diagnostics.visualize(profiler, file_path=profile_file, show=False, save=True)
                log.info(f'                 : wrote profiler info to {profile_file}')


    for (fld, spw, scan, antenna_or_baseline), df in dataframes.items():
        subset = index_subsets[fld, spw, scan, antenna_or_baseline]
        # update keys to be substituted into title and filename
        if fld is not None:
            keys['field_num'] = fld
            keys['field'] = ms.field[fld]
        if spw is not None:
            keys['spw'] = spw
        if scan is not None:
            keys['scan'] = scan
        if antenna_or_baseline is not None:
            if options.iter_ant:
                keys['ant'] = ms.all_antenna[antenna_or_baseline]
            elif options.iter_baseline:
                keys['baseline'] = ms.all_baseline[antenna_or_baseline]

        # now loop over plot types
        for props, xdatum, ydatum, adatum, ared, cdatum in all_plots:
            keys.update(title=props['title'], label=props['label'],
                        alphatitle=props['alpha_title'], alphalabel=props['alpha_label'],
                        colortitle=props['color_title'],
                        colorlabel=props['color_label'],
                        xname=xdatum.fullname, yname=ydatum.fullname,
                        xunit=xdatum.mapper.unit, yunit=ydatum.mapper.unit)
            if cdatum is not None and cdatum.is_discrete and not cdatum.discretized_labels:
                keys['colortitle'] += ' '+props['color_modulo']

            pngname = generate_string_from_keys(options.pngname, keys, "_", "_", "-")
            title   = generate_string_from_keys(options.title, keys, " ", " ", ", ")
            xlabel  = generate_string_from_keys(options.xlabel, keys, " ", " ", ", ")
            ylabel  = generate_string_from_keys(options.ylabel, keys, " ", " ", ", ")

            if options.dir:
                pngname = os.path.join(options.dir, pngname)

            # make output directory, if needed
            dirname = os.path.dirname(pngname)
            if dirname and not os.path.exists(dirname):
                os.mkdir(dirname)
                log.info(f'                 : created output directory {dirname}')

            if options.num_parallel < 2 or len(all_plots) < 2:
                render_single_plot(df, subset, xdatum, ydatum, adatum, ared, cdatum, pngname, title, xlabel, ylabel)
            else:
                from concurrent.futures import ThreadPoolExecutor
                executor = ThreadPoolExecutor(options.num_parallel)
                log.info(f'                 : submitting job for {pngname}')
                jobs.append(executor.submit(render_single_plot, df, subset, xdatum, ydatum, adatum, ared, cdatum,
                                            pngname, title, xlabel, ylabel))

    # wait for jobs to finish
    if executor is not None:
        log.info(f'                 : waiting for {len(jobs)} jobs to complete')
        for job in jobs:
            job.result()

    clock_stop = time.time()
    elapsed = str(round((clock_stop-clock_start), 2))

    log.info('Total time       : %s seconds' % (elapsed))

    if minmax_cache and options.lim_save:
        # ensure floats, because in64s and such cause errors
        minmax_cache = {axis: list(map(float, minmax)) for axis, minmax in minmax_cache.items()}

        with open(cache_file, "wt") as file:
            json.dump(minmax_cache, file, sort_keys=True, indent=4, separators=(',', ': '))
        log.info(f"Saved minmax cache to {cache_file} (disable with --no-lim-save)")

    log.info('Finished')
    separator()



# -*- coding: future_fstrings -*-

import argparse
import itertools
import numpy
import pkg_resources

from . import DEFAULT_CNUM, DEFAULT_NUM_RENDERS

try:
    __version__ = pkg_resources.require("shadems")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "dev"


def uppercase(str_):
    return str_.upper()


def rstrip_slash(str_):
    return str_.rstrip('/')


def cli():
    description = ("""Rapid Measurement Set plotting with dask-ms and datashader.
                   Version {0:s}"""
                   .format(__version__))
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("ms", type=rstrip_slash,  # remove / from ms by tab complete
        help="Measurement set")
    parser.add_argument("-v", "--version", action="version",
        version="{:s} version {:s}".format(parser.prog, __version__))

    group_opts = parser.add_argument_group("Plot types and data sources")
    group_opts.add_argument("-x", "--xaxis", action="append",
        help="""X axis of plot, e.g. "amp:CORRECTED_DATA"
             This recognizes all column names (also CHAN, FREQ, CORR, ROW, WAVEL, U, V, W, UV)
             and, for complex columns, keywords such as 'amp', 'phase', 'real', 'imag'.
             You can also specify correlations, e.g. 'DATA:phase:XX',
             and do two-column arithmetic with "+-*/", e.g. 'DATA-MODEL_DATA:amp'.
             Correlations may be specified by label, number, or as a Stokes parameter.
             The order of specifiers does not matter.""")
    group_opts.add_argument("-y", "--yaxis", action="append",
        help="""Y axis to plot.
             Must be given the same number of times as --xaxis.
             Note that X/Y can employ different columns and correlations.""")
    group_opts.add_argument("-a", "--aaxis", action="append",
        help="""Intensity axis.
             Can be none, or given once, or given the same number of times as --xaxis.
             If none, plot intensity (a.k.a. alpha channel) is proportional to density of points.
             Otherwise, a reduction function (see --ared below) is applied to the given values,
             and the result is used to determine intensity.
             All columns and variations listed under --xaxis are available for --aaxis.""")
    group_opts.add_argument("--ared", action="append",
        help="""Alpha axis reduction function.
             Recognized reductions are count, any, sum, min, max, mean, std, first, last, mode.
             Default is mean.""")

    group_opts.add_argument("-c", "--colour-by", action="append",
        help="""Colour (a.k.a. category) axis.
             Can be none, or given once, or given the same number of times as --xaxis.
             All columns and variations listed under --xaxis are available for colouring by.""")
    group_opts.add_argument("-C", "--col", metavar="COLUMN", dest="col", action="append", default=[],
        help="""Name of visibility column (default is DATA), if needed.
             This is used if the axis specifications do not explicitly include a column.
             For multiple plots, this can be given multiple times, or as a comma-separated list.
             Two-column arithmetic is recognized.""")

    group_opts.add_argument("--noflags", action="store_true",
        help="Enable to ignore flags. Default is to omit flagged data.")
    group_opts.add_argument("--noconj", action="store_true",
        help="Do not show conjugate points in u,v plots (default = plot conjugates).")

    group_opts = parser.add_argument_group("Plot axes setup")
    group_opts.add_argument("--xmin", action="append",
        help="""Minimum x-axis value (default = data min).
             For multiple plots, you can give this multiple times, or use a comma-separated list,
             but note that the clipping is the same per axis across all plots,
             so only the last applicable setting will be used.
             The list may include empty elements (or 'None') to not apply a clip.
             Default computes clips from data min/max.""")
    group_opts.add_argument("--xmax", action="append",
        help="Maximum x-axis value.")
    group_opts.add_argument("--ymin", action="append",
        help="Minimum y-axis value.")
    group_opts.add_argument("--ymax", action="append",
        help="Maximum y-axis value.")
    group_opts.add_argument("--amin", action="append",
        help="Minimum intensity-axis value.")
    group_opts.add_argument("--amax", action="append",
        help="Maximum intensity-axis value.")
    group_opts.add_argument("--cmin", action="append",
        help="Minimum value to be coloured by.")
    group_opts.add_argument("--cmax", action="append",
        help="Maximum value to be coloured by.")
    group_opts.add_argument("--cnum", action="append",
        help=f"""Number of steps used to discretize a continuous axis. Default is {DEFAULT_CNUM}.""")

    group_opts.add_argument("--xlim-load", action="store_true",
        help=f"Load x-axis limits from limits file, if available.")
    group_opts.add_argument("--ylim-load", action="store_true",
        help=f"Load y-axis limits from limits file, if available.")
    group_opts.add_argument("--clim-load", action="store_true",
        help=f"Load colour axis limits from limits file, if available.")
    group_opts.add_argument("--lim-file", default="{ms}-minmax-cache.json",
        help="""Name of limits file to save/load. '{ms}' will be substituted for MS base name.
             Default is '%(default)s'.""")
    group_opts.add_argument("--no-lim-save", action="store_false", dest="lim_save",
        help="""Do not save auto-computed limits to limits file. Default is to save.""")
    group_opts.add_argument("--lim-save-reset", action="store_true",
        help="""Reset limits file when saving. Default adds to existing file.""")

    group_opts = parser.add_argument_group("Options for multiple plots or combined plots")
    group_opts.add_argument("--iter-field", action="store_true",
        help="Separate plots per field (default is to combine in one plot)")
    group_opts.add_argument("--iter-spw", action="store_true",
        help="Separate plots per spw (default is to combine in one plot)")
    group_opts.add_argument("--iter-scan", action="store_true",
        help="Separate plots per scan (default is to combine in one plot)")
    group_opts.add_argument("--iter-corr", action="store_true",
        help="Separate plots per correlation or Stokes (default is to combine in one plot)")

    ex_group_opts = group_opts.add_mutually_exclusive_group()
    ex_group_opts.add_argument("--iter-ant", action="store_true",
        help="Separate plots per antenna (default is to combine in one plot)")
    ex_group_opts.add_argument("--iter-baseline", action="store_true",
        help="Separate plots per baseline (default is to combine in one plot)")

    group_opts = parser.add_argument_group("Data subset selection")
    ex_group_opts = group_opts.add_mutually_exclusive_group()
    ex_group_opts.add_argument("--ant", default="all",
        help="Antennas to plot (comma-separated list of names, default = all)")
    ex_group_opts.add_argument("--ant-num",
        help="""Antennas to plot (comma-separated list of numbers,
             or a [start]:[stop][:step] slice, mutually exclusive to --ant)""")
    group_opts.add_argument("--baseline", default="noautocorr",
        help="""Baselines to plot, as 'ant1-ant2' (comma-separated list:
             default of 'noautocorr' omits auto-correlations, 'autocorr' selects auto-correlations only, 'all' selects all baselines.
             Use 'ant1-*' to select all baselines to ant1.)""")
    group_opts.add_argument("--spw", default="all",
        help="""Spectral windows (DDIDs) to plot (comma-separated list, default = all)""")
    group_opts.add_argument("--field", default="all",
        help="Field ID(s) to plot (comma-separated list, default = all)")
    group_opts.add_argument("--scan", default="all",
        help="Scans to plot (comma-separated list, default = all)")
    group_opts.add_argument("--corr", default="all", type=uppercase,  # need this argument in upper case
        help="""Correlations or Stokes to plot, use indices or labels
             (comma-separated list, default = all)""")
    group_opts.add_argument("--chan",
        help="""Channel slice, as [start]:[stop][:step], default is to plot all channels""")

    group_opts = parser.add_argument_group("Rendering settings")
    group_opts.add_argument("-X", "--xcanvas", default=1280, type=int,
        help="Canvas x-size in pixels (default = %(default)s)")
    group_opts.add_argument("-Y", "--ycanvas", default=900, type=int,
        help="Canvas y-size in pixels (default = %(default)s)")
    group_opts.add_argument("--norm",
        choices=["auto", "eq_hist", "cbrt", "log", "linear"], default="auto",
        help="""Pixel scale normalization
             (default is 'log' with caxis, 'linear' with aaxis, and 'eq_hist' when neither is in use)""")
    group_opts.add_argument("--cmap", default="bkr",
        help="""Colorcet map used without --colour-by (default = %(default)s),
             see https://colorcet.holoviz.org""")
    group_opts.add_argument("--bmap", default="pride",
        help="""Colorcet map used when colouring by a continuous axis (default = %(default)s)""")
    group_opts.add_argument("--dmap", default="glasbey_dark",
        help="""Colorcet map used when colouring by a discrete axis (default = %(default)s)""")
    group_opts.add_argument("--dmap-preserve", action="store_true",
        help="""Preserve colour assignments in discrete axes even when discrete values are missing.""")
    group_opts.add_argument("--min-alpha", default=40, type=int, metavar="0-255",
        help="""Minimum alpha value used in rendering the canvas.
             Increase to saturate colour at the expense of dynamic range. Default is %(default)s.""")
    group_opts.add_argument("--saturate-perc", default=95, type=int, metavar="0-100",
        help="""Saturate colors so that the range [min-alpha, X] is mapped to [min-alpha, 255],
             where X is the given percentile. Default is %(default)s.""")
    group_opts.add_argument("--saturate-alpha", default=None, type=int, metavar="0-255",
        help="""Saturate colors as above, but with a fixed value of X. Overrides --saturate-perc.""")
    group_opts.add_argument("--spread-pix", type=int, default=0, metavar="PIX",
        help="Dynamically spread rendered pixels to this size")
    group_opts.add_argument("--spread-thr", type=float, default=0.5, metavar="THR",
        help="""Threshold parameter for spreading (0 to 1, default %(default)s)""")
    group_opts.add_argument("--bgcol", default="FFFFFF",
        help="RGB hex code for background colour (default = %(default)s)")
    group_opts.add_argument("--fontsize", default=16, type=float,
        help="Font size for all text elements (default = %(default)s)")

    group_opts.add_argument('--hline', type=str, metavar="Y[-[colour]]", 
                      help="Draw horizontal line(s) at given Y coordinate(s). You can append a matplotlib linestyle "
                           "(-, --, -., :) and/or a colour. You can also use a comma-separated list.")
    group_opts.add_argument('--vline', type=str, metavar="X[-[colour]]", 
                      help="Draw vertical line at given X coordinate(s). You can append a matplotlib linestyle "
                           "(-, --, -., :) and/or a colour. You can also use a comma-separated list.")
    group_opts.add_argument('-M', '--markup', type=str, action="append", nargs=2, metavar="func {args}", 
                      help="Add arbitrary matplotlib markup to plot. 'func' is a function known to matplotlib.Axes. "
                           "The {args} dict should be in YaML format.")

    group_opts = parser.add_argument_group("Output settings")
    # can also use "plot-{msbase}-{column}-{corr}-{xfullname}-vs-{yfullname}", let's expand on this later
    group_opts.add_argument("--dir",
        help="Send all plots to this output directory")
    group_opts.add_argument("-s", "--suffix",
        help="suffix to be included in filenames, can include {options}")
    group_opts.add_argument("--png", dest="pngname",
        default="plot-{ms}{_field}{_Spw}{_Scan}{_Ant}{_Baseline}-{label}{_alphalabel}{_colorlabel}{_suffix}.png",
        help='Template for output png files, default "%(default)s"')
    group_opts.add_argument("--title",
        default="{ms}{_field}{_Spw}{_Scan}{_Ant}{_Baseline}{_title}{_Alphatitle}{_Colortitle}",
        help='Template for plot titles, default "%(default)s"')
    group_opts.add_argument("--xlabel", default="{xname}{_xunit}",
        help='Template for X axis labels, default "%(default)s"')
    group_opts.add_argument("--ylabel", default="{yname}{_yunit}",
        help='Template for X axis labels, default "%(default)s"')

    optimization_opts = parser.add_argument_group("Performance & tweaking")
    optimization_opts.add_argument("-d", "--debug", action="store_true",
        help="Enable debugging output")
    optimization_opts.add_argument("-z", "--row-chunk-size", type=int, metavar="NROWS", default=5000,
        help="""Row chunk size for dask-ms.
             Larger chunks may or may not be faster, but will certainly use more RAM.""")
    optimization_opts.add_argument("-j", "--num-parallel", type=int, metavar="N", default=1,
        help=f"""Run up to N renderers in parallel.
             Default is serial.
             Use -j0 to auto-set this to half the available cores ({DEFAULT_NUM_RENDERS} on this system).
             This is not necessarily faster, as they might all end up contending for disk I/O.
             This might also work against dask-ms's own intrinsic parallelism.
             You have been advised.""")
    optimization_opts.add_argument("--profile", action="store_true",
        help="Enable dask profiling output")

    return parser, optimization_opts


# -- parser utility functions --
# flatten append lists generated by argparse.actions
flatten_list = lambda list_: list(itertools.chain(*[str_.split(',') for str_ in list_]))
unpack_axis = lambda opts_axis, default: [default] if not opts_axis else flatten_list(opts_axis)
# validate limits on axes, both min and max limits must be present
eval_none = lambda arr_1, arr_2: any([(a is None)^(b is None) for a, b in zip(arr_1, arr_2)])
# -- parser utility functions --


# -- dealing with input arguments to extract expected parameters --
def get_list(list_, type_=str, default=None, set_len=-1):
    """
        all parameter lists must be the same length as the expected number of plot
        and can also impose a type such as float (returning None for an empty string)

        parameters
        ----------
        list_: list
            appended list of input options
        type_: [optional] object
            desired variable type: e.g. int, float, str
            type_=str assumed if not specified
        default: [optional] object
            appropriate default value for variable
            default=None assumed if not specified
        set_len: [optional] int
            expected number of plots

        returns
        -------
        elems: list
            list of parameter options to plot,
            with a parameter for each plot to generate
    """
    # if empty list, set default
    if not list_:
        elems = [default]
    else:
        # stick all lists together
        elems = flatten_list(list_)
        # convert type
        elems = list(map(type_, elems))
    if len(elems) == 1 and set_len > 0:
        elems = [elems[0]]*set_len
    return elems


def parse_slice_spec(spec):
    """ check channel slice

        parameters
        ----------
        spec: str
            format [start]:[stop]:[step]

        returns
        -------
        slice: object
            python slice object
        elems: list
            slice element ([start, [stop, [step]]])
    """
    if spec:
        spec_elems = [int(x) if x else None for x in spec.split(":", 2)]
        return slice(*spec_elems), spec_elems
    else:
        return slice(None), []


def parse_plot_spec(parser, options):
    """
        unpacking the plotting input arguments into variables for plotting and
        making use the specified parameters makes sense for plotting

        parameters
        ----------
        parser: object
            raw ArgumentParser object
        options: object
            Namespace object containing parsed attributes

        returns
        -------
        xaxes: list
            xaxis parameter for each figure
        yaxes: list
            yaxis parameter for each figure
        columns: list
            name of visibility column per figure
        caxes: list
            color axis per figure (if specified, else None)
        aaxes: list
            blah, blah, blah, ...
        areds: list
            blah, blah, blah, ...
        xmins: list
            x-range minimum per figure (if specified, else None)
        xmaxs: list
            x-range maximum per figure (if specified, else None)
        ymins: list
            y-range minimum per figure (if specified, else None)
        ymaxs: list
            y-range maximum per figure (if specified, else None)
        amins: list
            blah, blah, blah, ...
        amaxs: list
            blah, blah, blah, ...
        cmins: list
            minimum colormap value per figure (if specified, else None)
        cmaxs: list
            maximum colormap value per figure (if specified, else None)
        cnums: list
            number of steps in colormap range
        chanslice: slice object (if channel selection is specified else All)
            easy selection of specified channels
    """

    # figure our list of plots to make
    # list of xaxis options to plot (default ['TIME'])
    xaxes = unpack_axis(options.xaxis, 'TIME')
    # list of yaxis associated with xaxes options (default['DATA:amp'])
    yaxes = unpack_axis(options.yaxis, 'DATA:amp')
    if len(xaxes) != len(yaxes):
        parser.error("--xaxis and --yaxis must be given the same number of times")

    # get list of columns and plot limits of the same length
    param_desc = {
                 'names': ('vars', 'opts', 'types', 'defaults'),
                 'formats': ('U15', 'U15', object, object)
                 }
    plot_params = numpy.array([
                              ('columns', 'col', str, ["DATA"]) ,
                              ('xmins', 'xmin', float, None),
                              ('xmaxs', 'xmax', float, None),
                              ('ymins', 'ymin', float, None),
                              ('ymaxs', 'ymax', float, None),
                              ('aaxes', 'aaxis', str, None),
                              ('amins', 'amin', float, None),
                              ('amaxs', 'amax', float, None),
                              ('areds', 'ared', str, 'mean'),
                              ('caxes', 'colour_by', str, None),
                              ('cmins', 'cmin', float, None),
                              ('cmaxs', 'cmax', float, None),
                              ('cnums', 'cnum', int, DEFAULT_CNUM),
                              ], dtype=param_desc)
    if not options.col:
        options.col = ["DATA"]
    # For all other settings, returns list same length as xaxes,
    # or throws error if no conformance.
    for param, name, dtype, default in zip(plot_params['vars'],
                                           plot_params['opts'],
                                           plot_params['types'],
                                           plot_params['defaults']):
        elems =  get_list(getattr(options, name, None),
                          type_=dtype,
                          default=default,
                          set_len=len(xaxes),
                          )
        if len(elems) > 1 and len(elems) != len(xaxes):
            parser.error(f"--{name} must be given the same number of times as --xaxis, or else just once")
        #variable is dynamically created and assigned
        globals()[param] = elems

    # check min/max
    if eval_none(xmins, xmaxs):
        parser.error("--xmin/--xmax must be either both set, or neither")
    if eval_none(ymins, ymaxs):
        parser.error("--ymin/--ymax must be either both set, or neither")
    if eval_none(cmins, cmaxs):
        parser.error("--cmin/--cmax must be either both set, or neither")
    if eval_none(amins, amaxs):
        parser.error("--amin/--amax must be either both set, or neither")

    return [xaxes, yaxes, columns, caxes, aaxes, areds,
            xmins, xmaxs, ymins, ymaxs, amins, amaxs, cmins, cmaxs, cnums]

# -fin-



import dask
import dask.array as da
import dask.dataframe as dd
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from shade_ms.dask_utils import dataframe_factory


@pytest.mark.parametrize("test_nan_shapes", [True, False])
def test_dataframe_factory(test_nan_shapes):
    nrow, nfreq = 100, 100

    data1a = da.arange(nrow, chunks=(10,))

    # Generate nan chunk shapes in data1a if requested
    if test_nan_shapes:
        data1a = data1a[da.where(data1a > 4)]

    data1b = da.random.random(size=(nfreq,), chunks=(100,))

    df = dataframe_factory(("row", "chan"),
                           data1a, ("row",),
                           data1b, ("chan",))

    assert isinstance(df, dd.DataFrame)
    assert isinstance(df['x'], dd.Series)
    assert isinstance(df['y'], dd.Series)

    if test_nan_shapes:
        # With unknown shapes, we broadcast our arrays in numpy
        # to test
        x, y = dask.compute(data1a[:, None], data1b[None, :])
        x, y = np.broadcast_arrays(x, y)
        x = x.ravel()
        y = y.ravel()

        # Unknown divisions
        assert df.divisions == (None,) * (df.npartitions + 1)
    else:
        # With know chunks we can broadcast our arrays in dask
        x, y = da.broadcast_arrays(data1a[:, None], data1b[None, :])
        x = x.ravel()
        y = y.ravel()

        # Known divisions
        assert df.divisions == (0, 1000, 2000, 3000, 4000, 5000,
                                6000, 7000, 8000, 9000, 10000)
        assert df['x'].npartitions == x.npartitions
        assert df['y'].npartitions == y.npartitions

    # Compare our lazy dataframe series vs (dask or numpy) arrays
    assert_array_equal(df['x'], x)
    assert_array_equal(df['y'], y)
    assert_array_equal(df['x'].min(), data1a.min())
    assert_array_equal(df['y'].min(), data1b.min())
    assert_array_equal(df['x'].max(), data1a.max())
    assert_array_equal(df['y'].max(), data1b.max())


def test_dataframe_factory_multicol():
    nrow, nfreq, ncorr = 100, 100, 4

    data1a = da.random.random(size=nrow, chunks=(10,))
    data1b = da.random.random(size=(nfreq, ncorr), chunks=(100, 4))
    data1c = da.random.random(size=(ncorr,), chunks=(4,))

    df = dataframe_factory(("row", "chan", "corr"),
                           data1a, ("row",),
                           data1b, ("chan", "corr"),
                           data1c, ("corr",))

    assert isinstance(df, dd.DataFrame)
    assert isinstance(df['x'], dd.Series)
    assert isinstance(df['y'], dd.Series)
    assert isinstance(df['c0'], dd.Series)


    x, y, c0 = da.broadcast_arrays(data1a[:, None, None],
                                   data1b[None, :, :],
                                   data1c[None, None, :])

    assert_array_equal(df['x'], x.ravel())
    assert_array_equal(df['y'], y.ravel())
    assert_array_equal(df['c0'], c0.ravel())

    assert_array_equal(df['x'].min(), data1a.min())
    assert_array_equal(df['x'].max(), data1a.max())
    assert_array_equal(df['y'].min(), data1b.min())
    assert_array_equal(df['y'].max(), data1b.max())
    assert_array_equal(df['c0'].min(), data1c.min())
    assert_array_equal(df['c0'].max(), data1c.max())

    df = df.append(df)

    assert_array_equal(df['x'].min(), data1a.min())
    assert_array_equal(df['x'].max(), data1a.max())
    assert_array_equal(df['y'].min(), data1b.min())
    assert_array_equal(df['y'].max(), data1b.max())
    assert_array_equal(df['c0'].min(), data1c.min())
    assert_array_equal(df['c0'].max(), data1c.max())



# -*- coding: future_fstrings -*-

import datashape
import datashape.coretypes as ct
import datashader.reductions
import numpy as np

try:
    import cudf
except ImportError:
    cudf = None

# couldn't find anything in the datashape docs about how to check if a CType is an integer,
# so just define a big set
IntegerTypes = {ct.bool_, ct.uint8, ct.uint16, ct.uint32, ct.uint64, ct.int8, ct.int16, ct.int32, ct.int64}


def _column_modulo(df, column, modulo):
    """
    Helper function. Takes a DataFrame column, modulo integer value
    """
    if cudf and isinstance(df, cudf.DataFrame):
        ## dunno how to do this in CUDA
        raise NotImplementedError("this feature is not implemented in cudf")
        ## but is it as simple as this?
        # return (df[column] % modulo).to_gpu_array()
    else:
        return df[column].values % modulo


def _column_discretization(df, column, offset, delta, nsteps):
    """
    Helper function. Takes a DataFrame column, modulo integer value
    """
    if cudf and isinstance(df, cudf.DataFrame):
        ## dunno how to do this in CUDA
        raise NotImplementedError("this feature is not implemented in cudf")
    else:
        value = df[column].values
        index = ((value - offset) / delta).astype(np.uint32)
        index[index < 0] = 0
        index[index >= nsteps] = nsteps - 1
        index[np.isnan(value)] = nsteps
        return index


class integer_modulo(datashader.reductions.category_codes):
    """A variation on category_codes that replaces categories by the values from an integer column, modulo a certain number"""
    def __init__(self, column, modulo):
        super().__init__(column)
        self.modulo = modulo

    def apply(self, df):
        return _column_modulo(df, self.column, self.modulo)


class float_discretization(integer_modulo):
    """A variation on category_codes that replaces categories by the values from an integer column, modulo a certain number"""
    def __init__(self, column, offset, delta, nsteps):
        super().__init__(column, nsteps)
        self.offset = offset
        self.delta = delta
        self.nsteps = nsteps

    def apply(self, df):
        return _column_discretization(df, self.column, self.offset, self.delta, self.nsteps)


class integer_modulo_values(datashader.reductions.category_values):
    """A variation on category_values that replaces categories by the values from an integer column, modulo a certain number"""
    def __init__(self, columns, modulo):
        super().__init__(columns)
        self.modulo = modulo

    def apply(self, df):
        a = _column_modulo(df, self.columns[0], self.modulo)
        return self._attach_value(df, a)

    def _attach_value(self, df, a):
        if cudf and isinstance(df, cudf.DataFrame):
            import cupy
            if df[self.columns[1]].dtype.kind == 'f':
                nullval = np.nan
            else:
                nullval = 0
            b = df[self.columns[1]].to_gpu_array(fillna=nullval)
            return cupy.stack((a, b), axis=-1)
        else:
            b = df[self.columns[1]].values
            return np.stack((a, b), axis=-1)


class float_discretization_values(integer_modulo_values):
    """A variation on category_codes that replaces categories by the values from an integer column, modulo a certain number"""
    def __init__(self, columns, offset, delta, nsteps):
        super().__init__(columns, nsteps+1)  # one extra category for NaNs
        self.offset = offset
        self.delta = delta
        self.nsteps = nsteps

    def apply(self, df):
        a = _column_discretization(df, self.columns[0], self.offset, self.delta, self.nsteps)
        return self._attach_value(df, a)


class by_integers(datashader.by):
    """Like datashader.by, but for integer-valued columns."""
    def __init__(self, cat_column, reduction, modulo):
        super().__init__(cat_column, reduction)
        self.modulo = modulo

    def _build_temps(self, cuda=False):
        return tuple(by_integers(self.cat_column, tmp, self.modulo) for tmp in self.reduction._build_temps(cuda))

    def validate(self, in_dshape):
        if not self.cat_column in in_dshape.dict:
            raise ValueError("specified column not found")
        if not in_dshape.measure[self.cat_column] in IntegerTypes:
            raise ValueError("input must be an integer column")
        self.reduction.validate(in_dshape)

    def out_dshape(self, input_dshape):
        cats = list(range(self.modulo))
        red_shape = self.reduction.out_dshape(input_dshape)
        return datashape.util.dshape(datashape.Record([(c, red_shape) for c in cats]))

    @property
    def inputs(self):
        if self.val_column is not None:
            return (integer_modulo_values(self.columns, self.modulo),)
        else:
            return (integer_modulo(self.columns[0], self.modulo),)

    def _build_bases(self, cuda=False):
        bases = self.reduction._build_bases(cuda)
        if len(bases) == 1 and bases[0] is self:
            return bases
        return tuple(by_integers(self.cat_column, base, self.modulo) for base in bases)

    def _build_finalize(self, dshape):
        cats = list(range(self.modulo))

        def finalize(bases, cuda=False, **kwargs):
            kwargs['dims'] += [self.cat_column]
            kwargs['coords'][self.cat_column] = cats
            return self.reduction._finalize(bases, cuda=cuda, **kwargs)

        return finalize

class by_span(by_integers):
    """Like datashader.by, but for float-valued columns, discretized by steps over a certain span"""
    def __init__(self, cat_column, reduction, offset, delta, nsteps):
        super().__init__(cat_column, reduction, nsteps+1)  # allocate extra category for NaNs
        self.offset = offset
        self.delta = delta
        self.nsteps = nsteps

    def _build_temps(self, cuda=False):
        return tuple(by_span(self.cat_column, tmp, self.offset, self.delta, self.nsteps)
                     for tmp in self.reduction._build_temps(cuda))

    def validate(self, in_dshape):
        if not self.cat_column in in_dshape.dict:
            raise ValueError("specified column not found")
        self.reduction.validate(in_dshape)

    @property
    def inputs(self):
        if self.val_column is not None:
            return (float_discretization_values(self.columns, self.offset, self.delta, self.nsteps),)
        else:
            return (float_discretization(self.columns[0], self.offset, self.delta, self.nsteps),)

    def _build_bases(self, cuda=False):
        bases = self.reduction._build_bases(cuda)
        if len(bases) == 1 and bases[0] is self:
            return bases
        return tuple(by_span(self.cat_column, base, self.offset, self.delta, self.nsteps) for base in bases)




from MSUtils.msutils import STOKES_TYPES
from casacore.tables import table
import re
import daskms
import math
import numpy as np
from collections import OrderedDict

class NamedList(object):
    """Holds a list of names (e.g. field names), and provides common indexing and subset operations"""

    def __init__(self, label, names, numbers=None):
        self.label = label
        self.names = names
        self.numbers = numbers or range(len(self.names))
        self.map = dict(zip(names, self.numbers))
        self.numindex = {num: i for i, num in enumerate(self.numbers)}

    def __len__(self):
        return len(self.names)

    def __contains__(self, name):
        return name in self.map if type(name) is str else name in self.numbers

    def __getitem__(self, item, default=None):
        if type(item) is str:
            return self.map.get(item, default)
        elif type(item) is slice:
            return self.names[item]
        else:
            return self.names[item] # self.numindex[item]]

    def get_subset(self, subset, allow_numeric_indices=True):
        """Extracts subset using a comma-separated string or list of indices"""
        if type(subset) in (list, tuple):
            return NamedList(self.label, [self.names[x] for x in subset], [self.numbers[x] for x in subset])
        elif type(subset) is str:
            if subset == "all":
                return self
            ind = []
            for x in subset.split(","):
                if allow_numeric_indices and re.fullmatch('\d+', x):
                    x = int(x)
                    if x not in self.numindex:
                        raise ValueError(f"invalid {self.label} number {x}")
                    ind.append(self.numindex[x])
                elif x in self.map:
                    ind.append(self.numindex[self.map[x]])
                else:
                    raise ValueError(f"invalid {self.label} '{x}'")
            return NamedList(self.label, [self.names[x] for x in ind], [self.numbers[x] for x in ind])
        else:
            raise TypeError(f"unknown subset of type {type(subset)}")

    def str_list(self):
        return " ".join([f"{i}:{name}" for i, name in zip(self.numbers, self.names)])

class MSInfo(object):
    """Holds information about the MS structure"""

    def __init__(self, msname=None, log=None):
        if not msname:
            return

        self.msname = msname
        self.log = log

        tab = table(msname, ack=False)
        log and log.info(f": MS {msname} contains {tab.nrows()} rows")

        self.valid_columns = set(tab.colnames())

        # indexing columns are read into memory in their entirety up front
        self.indexing_columns = dict()

        spw_tab = daskms.xds_from_table(msname + '::SPECTRAL_WINDOW', columns=['CHAN_FREQ'])
        self.chan_freqs = spw_tab[0].CHAN_FREQ   # important for this to be an xarray
        self.nspw = self.chan_freqs.shape[0]
        self.spw = NamedList("spw", list(map(str, range(self.nspw))))

        log and log.info(f":   {self.chan_freqs.shape} spectral windows and channels")

        self.field = NamedList("field", table(msname +'::FIELD', ack=False).getcol("NAME"))
        log and log.info(f":   {len(self.field)} fields: {' '.join(self.field.names)}")

        self.indexing_columns["SCAN_NUMBER"] = tab.getcol("SCAN_NUMBER")
        scan_numbers = sorted(set(self.indexing_columns["SCAN_NUMBER"]))
        log and log.info(f":   {len(scan_numbers)} scans: {' '.join(map(str, scan_numbers))}")
        all_scans = NamedList("scan", list(map(str, range(scan_numbers[-1]+1))))
        self.scan = all_scans.get_subset(scan_numbers)

        anttab = table(msname +'::ANTENNA', ack=False)
        antnames = anttab.getcol("NAME")
        self.all_antenna = antnames = NamedList("antenna", antnames)
        self.antpos = anttab.getcol("POSITION")

        ant1col = self.indexing_columns["ANTENNA1"] = tab.getcol("ANTENNA1")
        ant2col = self.indexing_columns["ANTENNA2"] = tab.getcol("ANTENNA2")
        self.antenna = self.all_antenna.get_subset(list(set(ant1col)|set(ant2col)))

        log and log.info(f":   {len(self.antenna)}/{len(self.all_antenna)} antennas: {self.antenna.str_list()}")

        # list of all possible baselines
        nant = len(antnames)
        blnames = [f"{a1}-{a2}" for i1, a1 in enumerate(antnames) for a2 in antnames[i1:]]
        self.all_baseline = NamedList("baseline", blnames)

        # list of baseline lengths
        self.baseline_lengths = [math.sqrt(((pos1-pos2)**2).sum())
                                 for i1, pos1 in enumerate(self.antpos) for pos2 in self.antpos[i1:]]

        # sort order to put baselines by length
        sorted_by_length = sorted([(x, i) for i, x in enumerate(self.baseline_lengths)])
        self.baseline_sorted_index  = [i for _, i in sorted_by_length]
        self.baseline_sorted_length = [x for x, _ in sorted_by_length]

        # baselines actually present
        a1 = np.minimum(ant1col, ant2col)
        a2 = np.maximum(ant1col, ant2col)
        bl_set = set(self.baseline_number(a1, a2))
        bls = sorted(bl_set)
        self.baseline = NamedList("baseline", [blnames[b] for b in bls], bls)

        # make list of baselines present, in meters
        blm = [i for i in self.baseline_sorted_index if i in bl_set]
        self.baseline_m = NamedList("baseline_m", [f"{int(round(self.baseline_lengths[b]))}m" for b in blm], blm)

        log and log.info(f":   {len(self.baseline)}/{len(self.all_baseline)} baselines present")

        pol_tab = table(msname + '::POLARIZATION', ack=False)

        all_corr_labels = [STOKES_TYPES[icorr] for icorr in pol_tab.getcol("CORR_TYPE", 0, 1).ravel()]
        self.corr = NamedList("correlation", all_corr_labels.copy())

        # Maps correlation -> callable that extracts that correlation from visibility data
        # By default, populated with slicing functions for 0...3,
        # but can also be extended with "I", "Q", etx.
        self.corr_data_mappers = OrderedDict({i: lambda x,icorr=i:x[...,icorr] for i in range(len(all_corr_labels))})

        # Maps correlation -> callable that extracts that correlation from flag data
        self.corr_flag_mappers = self.corr_data_mappers.copy()

        # add mappings and labels for Stokes parameters
        xx, xy, yx, yy = [self.corr.map.get(c) for c in ("XX", "XY", "YX", "YY")]
        rr, rl, lr, ll = [self.corr.map.get(c) for c in ("RR", "RL", "LR", "LL")]

        def add_stokes(a, b, I, J, imag=False):
            """Adds mappers for Stokes A and B as the sum/difference of components I and J, divided by 2 or 2j"""
            def _sum(x):
                return (x[..., I] + x[..., J]) / 2
            def _diff(x):
                return (x[..., I] - x[..., J]) / (2j if imag else 2)
            def _or(x):
                return (x[..., I] | x[..., J])
            nonlocal all_corr_labels
            if a not in self.corr_data_mappers:
                self.corr_data_mappers[len(all_corr_labels)] = _sum
                self.corr_flag_mappers[len(all_corr_labels)] = _or
                all_corr_labels.append(a)
            if b not in self.corr_data_mappers:
                self.corr_data_mappers[len(all_corr_labels)] = _diff
                self.corr_flag_mappers[len(all_corr_labels)] = _or
                all_corr_labels.append(b)

        if xx is not None and yy is not None:
            add_stokes("I", "Q", xx, yy)
        if rr is not None and ll is not None:
            add_stokes("I", "V", rr, ll)
        if xy is not None and yx is not None:
            add_stokes("U", "V", xy, yx, True)
        if rl is not None and lr is not None:
            add_stokes("Q", "U", rl, lr, True)

        self.all_corr = NamedList("correlation", all_corr_labels)

        log and log.info(f":   corrs/Stokes {' '.join(self.all_corr.names)}")

    def baseline_number(self, a1, a2):
        """Returns baseline number, for a1<=a2"""
        return a1 * len(self.all_antenna) - a1 * (a1 - 1) // 2 + a2 - a1




# -*- coding: future_fstrings -*-

# ian.heywood@physics.ox.ac.uk

import matplotlib
matplotlib.use('agg')

import daskms
import dask.array as da
import dask.dataframe as dask_df
import xarray
import holoviews as holoviews
import holoviews.operation.datashader
import datashader.transfer_functions
import datashader.reductions
from datashader.reductions import category_modulo, category_binning
import numpy as np
import pylab
import textwrap
import argparse
import itertools
import matplotlib.cm
from shade_ms import log
import colorcet
import cmasher
import matplotlib.cm

from collections import OrderedDict
from . import data_mappers
from .data_mappers import DataAxis
from .dask_utils import dataframe_factory
# from .ds_ext import by_integers, by_span

def add_options(parser):
    pass


def set_options(options):
    pass


def get_colormap(cmap_name):
    cmap = getattr(colorcet, cmap_name, None)
    if cmap:
        log.info(f"using colourmap colorcet.{cmap_name}")
        return cmap
    cmap = getattr(cmasher, cmap_name, None)
    if cmap:
        log.info(f"using colourmap cmasher.{cmap_name}")
    else:
        cmap = getattr(matplotlib.cm, cmap_name, None)
        if cmap is None:
            raise ValueError(f"unknown colourmap {cmap_name}")
        log.info(f"using colourmap matplotplib.cm.{cmap_name}")
    return [ f"#{int(r*255):02x}{int(g*255):02x}{int(b*255):02x}" for r,g,b in cmap.colors ]


def freq_to_wavel(ff):
    c = 299792458.0  # m/s
    return c/ff

def get_plot_data(msinfo, group_cols, mytaql, chan_freqs,
                  chanslice, subset,
                  noflags, noconj,
                  iter_field, iter_spw, iter_scan, iter_ant, iter_baseline,
                  join_corrs=False,
                  row_chunk_size=100000):

    ms_cols = {'ANTENNA1', 'ANTENNA2'}
    ms_cols.update(msinfo.indexing_columns.keys())
    if not noflags:
        ms_cols.update({'FLAG', 'FLAG_ROW'})
    # get visibility columns
    for axis in DataAxis.all_axes.values():
        ms_cols.update(axis.columns)

    total_num_points = 0  # total number of points to plot

    # output dataframes, indexed by (field, spw, scan, antenna_or_baseline)
    # If any of these axes is not being iterated over, then the index at that position is None
    output_dataframes = OrderedDict()

    # number of rows per each dataframe
    output_rows = OrderedDict()

    # output subsets of indexing columns, indexed by same tuple
    output_subsets = OrderedDict()

    if iter_ant:
        antenna_subsets = zip(subset.ant.numbers, subset.ant.names)
    else:
        antenna_subsets = [(None, None)]
    taql = mytaql

    for antenna, antname in antenna_subsets:
        if antenna is not None:
            taql = f"({mytaql})&&(ANTENNA1=={antenna} || ANTENNA2=={antenna})" if mytaql else \
                    f"(ANTENNA1=={antenna} || ANTENNA2=={antenna})"
        # add baselines to group columns
        if iter_baseline:
            group_cols = list(group_cols) + ["ANTENNA1", "ANTENNA2"]

        # get MS data
        msdata = daskms.xds_from_ms(msinfo.msname, columns=list(ms_cols), group_cols=group_cols, taql_where=taql,
                                    chunks=dict(row=row_chunk_size))
        nrow = sum([len(group.row) for group in msdata])
        if not nrow:
            continue

        if antenna is not None:
            log.info(f': Indexing sub-MS (antenna {antname}) and building dataframes ({nrow} rows, chunk size is {row_chunk_size})')
        else:
            log.info(f': Indexing MS and building dataframes ({nrow} rows, chunk size is {row_chunk_size})')

        # iterate over groups
        for group in msdata:
            if not len(group.row):
                continue
            ddid     =  group.DATA_DESC_ID  # always present
            fld      =  group.FIELD_ID      # always present
            if fld not in subset.field or ddid not in subset.spw:
                log.debug(f"field {fld} ddid {ddid} not in selection, skipping")
                continue
            scan = getattr(group, 'SCAN_NUMBER', None)  # will be present if iterating over scans
            if iter_baseline:
                ant1    = getattr(group, 'ANTENNA1', None)   # will be present if iterating over baselines
                ant2    = getattr(group, 'ANTENNA2', None)   # will be present if iterating over baselines
                baseline = msinfo.baseline_number(ant1, ant2)
            else:
                baseline = None

            # Make frame key -- data subset corresponds to this frame
            dataframe_key = (fld if iter_field else None,
                             ddid if iter_spw else None,
                             scan if iter_scan else None,
                             antenna if antenna is not None else baseline)

            # update subsets of MS indexing columns that we've seen for this dataframe
            output_subset1 = output_subsets.setdefault(dataframe_key,
                                                {column:set() for column in msinfo.indexing_columns.keys()})
            for column, _ in msinfo.indexing_columns.items():
                value = getattr(group, column)
                if np.isscalar(value):
                    output_subset1[column].add(value)
                else:
                    output_subset1[column].update(value.compute().data)

            # number of rows in dataframe
            nrows0 = output_rows.setdefault(dataframe_key, 0)

            # always read flags -- easier that way
            flag = group.FLAG if not noflags else None
            flag_row = group.FLAG_ROW if not noflags else None

            a1 = da.minimum(group.ANTENNA1.data, group.ANTENNA2.data)
            a2 = da.maximum(group.ANTENNA1.data, group.ANTENNA2.data)
            baselines = msinfo.baseline_number(a1, a2)

            freqs = chan_freqs[ddid]
            chans = xarray.DataArray(range(len(freqs)), dims=("chan",))
            wavel = freq_to_wavel(freqs)
            extras = dict(chans=chans, freqs=freqs, wavel=wavel, rows=group.row, baselines=baselines)

            nchan = len(group.chan)
            if flag is not None:
                flag = flag[dict(chan=chanslice)]
                nchan = flag.shape[1]
            shape = (len(group.row), nchan)

            arrays = OrderedDict()
            shapes = OrderedDict()
            ddf = None
            num_points = 0  # counts number of new points generated


            for corr in subset.corr.numbers:
                # make dictionary of extra values for DataMappers
                extras['corr'] = corr
                # loop over datums to be computed
                for axis in DataAxis.all_axes.values():
                    value = arrays.get(axis.label)
                    # a datum was already computed?
                    if value is not None:
                        # if not joining correlations, then that's the only one we'll need, so continue
                        if not join_corrs:
                            continue
                        # joining correlations, and datum has a correlation dependence: compute another one
                        if axis.corr is None:
                            value = None
                    if value is None:
                        value = axis.get_value(group, corr, extras, flag=flag, flag_row=flag_row, chanslice=chanslice)
                        # print(axis.label, value.compute().min(), value.compute().max())
                        num_points = max(num_points, value.size)
                        if value.ndim == 0:
                            shapes[axis.label] = ()
                        elif value.ndim == 1:
                            timefreq_axis = axis.mapper.axis or 0
                            assert value.shape[0] == shape[timefreq_axis], \
                                   f"{axis.mapper.fullname}: size {value.shape[0]}, expected {shape[timefreq_axis]}"
                            shapes[axis.label] = ("row",) if timefreq_axis == 0 else ("chan",)
                        # else 2D value better match expected shape
                        else:
                            assert value.shape == shape, f"{axis.mapper.fullname}: shape {value.shape}, expected {shape}"
                            shapes[axis.label] = ("row", "chan")
                        arrays[axis.label] = value
                # any new data generated for this correlation? Make dataframe
                if num_points:
                    total_num_points += num_points
                    args = (v for pair in ((array, shapes[key]) for key, array in arrays.items()) for v in pair)
                    df1 = dataframe_factory(("row", "chan"), *args, columns=arrays.keys())
                    # if any axis needs to be conjugated, double up all of them
                    if not noconj and any([axis.conjugate for axis in DataAxis.all_axes.values()]):
                        arr_shape = [(-arrays[axis.label] if axis.conjugate else arrays[axis.label], shapes[axis.label])
                                                for axis in DataAxis.all_axes.values()]
                        args = (v for pair in arr_shape  for v in pair)
                        df2 = dataframe_factory(("row", "chan"), *args, columns=arrays.keys())
                        df1 = dask_df.concat([df1, df2], axis=0)
                    ddf = dask_df.concat([ddf, df1], axis=0) if ddf is not None else df1

            # do we already have a frame for this key
            ddf0 = output_dataframes.get(dataframe_key)

            if ddf0 is None:
                log.debug(f"first frame for {dataframe_key}")
                output_dataframes[dataframe_key] = ddf
            else:
                log.debug(f"appending to frame for {dataframe_key}")
                output_dataframes[dataframe_key] = dask_df.concat([ddf0, ddf], axis=0)

    # convert discrete axes into categoricals
    if data_mappers.USE_COUNT_CAT:
        categorical_axes = [axis.label for axis in DataAxis.all_axes.values() if axis.nlevels]
        if categorical_axes:
            log.info(": counting colours")
            for key, ddf in list(output_dataframes.items()):
                output_dataframes[key] = ddf.categorize(categorical_axes)

    # print("===")
    # for ddf in output_dataframes.values():
    #     for axis in DataAxis.all_axes.values():
    #         value = ddf[axis.label].values.compute()
    #         print(axis.label, np.nanmin(value), np.nanmax(value))

    log.info(": complete")
    return output_dataframes, output_subsets, total_num_points

def compute_bounds(unknowns, bounds, ddf):
    """
    Given a list of axis with unknown bounds, computes missing bounds and updates the bounds dict
    """
    # setup function to compute min/max on every column for which we don't have a min/max
    with np.errstate(all='ignore'):
        r = ddf.map_partitions(lambda df:
                np.array([[(np.nanmin(df[axis].values).item() if bounds[axis][0] is None else bounds[axis][0]) for axis in unknowns]+
                          [(np.nanmax(df[axis].values).item() if bounds[axis][1] is None else bounds[axis][1]) for axis in unknowns]]),
            ).compute()

    # setup new bounds dict based on this
    for i, axis in enumerate(unknowns):
        minval = np.nanmin(r[:, i])
        maxval = np.nanmax(r[:, i + len(unknowns)])
        if not (np.isfinite(minval) and np.isfinite(maxval)):
            minval, maxval = -1.0, 1.0
        elif minval >= maxval:
            minval, maxval = minval-1, minval+1
        bounds[axis] = minval, maxval



def create_plot(ddf, index_subsets, xdatum, ydatum, adatum, ared, cdatum, cmap, bmap, dmap, normalize,
                xlabel, ylabel, title, pngname,
                extra_markup=[],
                min_alpha=40, saturate_percentile=None, saturate_alpha=None,
                minmax_cache=None,
                options=None):

    figx = options.xcanvas / 60
    figy = options.ycanvas / 60
    bgcol = "#" + options.bgcol.lstrip("#")

    xaxis = xdatum.label
    yaxis = ydatum.label
    aaxis = adatum and adatum.label
    caxis = cdatum and cdatum.label

    color_key = color_labels = color_minmax = agg_alpha = None

    # do we need to compute any axis min/max?
    bounds = OrderedDict({xaxis: xdatum.minmax, yaxis: ydatum.minmax})
    unknown = []
    for datum in xdatum, ydatum, cdatum:
        if datum is not None:
            bounds[datum.label] = datum.minmax
            if datum.minmax[0] is None or datum.minmax[1] is None:
                if datum.is_discrete and datum.subset_indices is not None:
                    bounds[datum.label] = 0, len(datum.subset_indices)-1
                else:
                    unknown.append(datum.label)

    if unknown:
        log.info(f": scanning axis min/max for {' '.join(unknown)}")
        compute_bounds(unknown, bounds, ddf)
        # populate cache
        if minmax_cache is not None:
            minmax_cache.update([(label, bounds[label]) for label in unknown])

    # adjust bounds for discrete axes
    canvas_sizes = []
    for datum, size in (xdatum, options.xcanvas), (ydatum, options.ycanvas):
        if datum.is_discrete:
            bounds[datum.label] = bounds[datum.label][0]-0.5, bounds[datum.label][1]+0.5
            size = int(bounds[datum.label][1]) - int(bounds[datum.label][0]) + 1
        canvas_sizes.append(size)

    # create rendering canvas.
    canvas = datashader.Canvas(canvas_sizes[0], canvas_sizes[1], x_range=bounds[xaxis], y_range=bounds[yaxis])

    if aaxis is not None:
        agg_alpha = getattr(datashader.reductions, ared, None) if ared else datashader.reductions.count
        if agg_alpha is None:
            raise ValueError(f"unknown alpha reduction function {ared}")
        agg_alpha = agg_alpha(aaxis)

    if cdatum is not None:
        # aggregation applied to by()
        agg_by = agg_alpha if agg_alpha else datashader.count()

        # figure out mapping from raster planes to colours
        # after this if-else block, category will be an aggregator instance yielding N categories,
        # color_key will be a list of N colors, and color_label will be a list of N textual labels

        if data_mappers.USE_COUNT_CAT:
            cats = getattr(ddf.dtypes, caxis).categories
            log.debug(f'colourizing using {caxis} categorical, {len(cats)} bins')
            category = caxis
            color_key = dmap[:len(cats)]
            color_labels = list(map(str, cats))
        else:
            if cdatum.is_discrete:
                # make dictionary from index to label, omitting values that are not in the MS subset to begin with
                if cdatum.discretized_labels:
                    active_subset = OrderedDict(enumerate(cdatum.discretized_labels))
                # else make up integer labels on the spot
                else:
                    active_subset = OrderedDict(enumerate(map(str, range(bounds[caxis][1]+1))))
                # Check if the subset needs to be refined, because it is known to be smaller for this dataframe
                if len(cdatum.columns) == 1 and cdatum.columns[0] in index_subsets:
                    df_index_subset = index_subsets[cdatum.columns[0]]
                    if cdatum.subset_remapper is not None:
                        remapper = cdatum.subset_remapper.compute()
                        df_index_subset = set(remapper[x] for x in df_index_subset)
                    active_subset = OrderedDict((idx, active_subset[idx]) for idx in df_index_subset)
                    log.debug(f"subset of indices for this axis is a priori {list(active_subset.keys())}")
                # max known index
                max_index = max(active_subset.keys())
                num_colors = min(cdatum.nlevels, len(dmap))
                color_key = dmap[:num_colors]
                # if we have fewer indices than colour levels, and the max index is sensible, we'll aggregate to one
                # raster slice per index value directly
                if len(active_subset) <= num_colors and max_index < max(num_colors, 256):
                    num_colors = max_index+1
                    log.debug(f"aggregating directly into {max_index+1} categories")
                    category = category_modulo(caxis, max_index+1)
                    color_label_list = {idx: [value] for idx, value in active_subset.items()}
                else:
                    log.debug(f"aggregating modulo {num_colors} categories")
                    category = category_modulo(caxis, num_colors)
                    # each slice maps to, potentially, multiple labels from the subset
                    color_label_list = {i: [active_subset[idx] for idx in range(i, max_index+1, num_colors) if idx in active_subset]
                                        for i in range(num_colors)}
                    # and colors just come from the bottom of the colormap
                    color_dict = dict(enumerate(options.dmap[:num_colors]))
                # convert lists of color labels into strings
                color_labels = ['']*num_colors
                for i, labels in color_label_list.items():
                    if len(labels) < 3:
                        color_labels[i] = ",".join(labels)
                    else:
                        color_labels[i] = ",".join(labels[:2] + ["..."])
            # else we discretize a span of values
            else:
                num_colors = min(cdatum.nlevels, len(bmap))
                log.debug(f'colourizing using {caxis} with {num_colors} bins')
                cmin = bounds[caxis][0]
                cdelta = (bounds[caxis][1] - cmin) / num_colors
                category = category_binning(caxis, cmin, cdelta, num_colors)
                # color labels are bin centres
                bin_centers = [cmin + cdelta*(i+0.5) for i in range(num_colors)]
                # map to colors pulled from entire extent of color map
                color_key = [bmap[(i*len(bmap))//num_colors] for i in range(num_colors)]
                color_labels = [str(bin) for bin in bin_centers]
                log.info(f": aggregating using {num_colors} bins at {' '.join(color_labels)})")

        raster = canvas.points(ddf, xaxis, yaxis, agg=datashader.by(category, agg_by))
        is_integer_raster = np.issubdtype(raster.dtype, np.integer)

        # the binning aggregator accumulates flagged points in an extra raster plane
        if isinstance(category, category_binning):
            if is_integer_raster:
                log.info(f": {raster[..., -1].data.sum():.3g} points were flagged ")
            raster = raster[...,:-1]

        if is_integer_raster:
            non_empty = np.array(raster.any(axis=(0, 1)))
        else:
            non_empty = ~(np.isnan(raster.data).all(axis=(0, 1)))
        if not non_empty.any():
            log.info(": no valid data in plot. Check your flags and/or plot limits.")
            return None

        if cdatum.is_discrete and not data_mappers.USE_COUNT_CAT:
            # discard empty planes
            non_empty = np.where(non_empty)[0]
            raster = raster[..., non_empty]
            # compress colours to bottom of colormap, unless asked to preserve assignments
            if options.dmap_preserve:
                color_key = [color_key[bin] for bin in non_empty]
            else:
                color_key = color_key[:len(non_empty)]
            color_labels =  [color_labels[bin] for bin in non_empty]

        img = datashader.transfer_functions.shade(raster, color_key=color_key, how=normalize, min_alpha=min_alpha)
        # set color_minmax for colorbar
        color_minmax = bounds[caxis]
    else:
        log.debug(f'rasterizing using {ared}')
        raster = canvas.points(ddf, xaxis, yaxis, agg=agg_alpha)
        if not raster.data.any():
            log.info(": no valid data in plot. Check your flags and/or plot limits.")
            return None
        # get min/max cor colorbar
        if aaxis:
            amin, amax = adatum.minmax
            color_minmax = (amin if amin is not None else np.nanmin(raster)), \
                           (amax if amax is not None else np.nanmax(raster))
            color_key = cmap
        log.debug('shading')
        img = datashader.transfer_functions.shade(raster, cmap=cmap, how=normalize, span=color_minmax, min_alpha=min_alpha)

    # resaturate if needed
    if saturate_alpha is not None or saturate_percentile is not None:
        # get alpha channel
        imgval = img.values
        alpha = (imgval >> 24)&255
        nulls = alpha<min_alpha
        alpha -= min_alpha
        if nulls.all():
            log.debug(f"alpha<min_alpha for entire plot -- all data below lower clip perhaps?")
        else:
            #if percentile if specified, use that to override saturate_alpha
            if saturate_alpha is None:
                saturate_alpha = np.percentile(alpha[~nulls], saturate_percentile)
                log.debug(f"using saturation alpha {saturate_alpha} from {saturate_percentile}th percentile")
            else:
                log.debug(f"using explicit saturation alpha {saturate_alpha}")
            # rescale alpha from [min_alpha, saturation_alpha] to [min_alpha, 255]
            saturation_factor = (255. - min_alpha) / (saturate_alpha - min_alpha)
            alpha = min_alpha + alpha*saturation_factor
            alpha[nulls] = 0
            alpha[alpha>255] = 255
            imgval[:] = (imgval & 0xFFFFFF) | alpha.astype(np.uint32)<<24

    if options.spread_pix:
        img = datashader.transfer_functions.dynspread(img, options.spread_thr, max_px=options.spread_pix)
        log.info(f": spreading ({options.spread_thr} {options.spread_pix})")
    rgb = holoviews.RGB(holoviews.operation.datashader.shade.uint32_to_uint8_xr(img))

    log.debug('done')

    # Set plot limits based on data extent or user values for axis labels

    xmin, xmax = bounds[xaxis]
    ymin, ymax = bounds[yaxis]

    log.debug('rendering image')

    fig = pylab.figure(figsize=(figx, figy))
    ax = fig.add_subplot(111, facecolor=bgcol)

    for funcname, args, kwargs in extra_markup:
        getattr(ax, funcname)(*args, **kwargs)

    ax.imshow(X=rgb.data, extent=[xmin, xmax, ymin, ymax],
              aspect='auto', origin='lower', interpolation='nearest')

    ax.set_title("\n".join(textwrap.wrap(title, 90)), loc='center', fontdict=dict(fontsize=options.fontsize))
    ax.set_xlabel(xlabel, fontdict=dict(fontsize=options.fontsize))
    ax.set_ylabel(ylabel, fontdict=dict(fontsize=options.fontsize))
    # ax.plot(xmin,ymin,'.',alpha=0.0)
    # ax.plot(xmax,ymax,'.',alpha=0.0)

    dx, dy = xmax - xmin, ymax - ymin
    ax.set_xlim([xmin - dx/100, xmax + dx/100])
    ax.set_ylim([ymin - dy/100, ymax + dy/100])

    def decimate_list(x, maxel):
        """Helper function to reduce a list to < given max number of elements, dividing it by decimal factors of 2 and 5"""
        factors = 2, 5, 10
        base = divisor = 1
        while len(x)//divisor > maxel:
            for fac in factors:
                divisor = fac*base
                if len(x)//divisor <= maxel:
                    break
            base *= 10
        return x[::divisor]

    ax.tick_params(labelsize=options.fontsize*0.66)

    # max # of tickmarks and labels to draw for discrete axes
    MAXLABELS = 64   # if we have up to this many labels, show them all
    MAXLABELS1 = 32  # if we have >MAXLABELS to show, then sparsify and get below this number
    MAXTICKS = 300   # if total number of points is within this range, draw them as minor tickmarks

    # do we have discrete labels to put on the axes?
    if xdatum.discretized_labels is not None:
        n = len(xdatum.discretized_labels)
        ticks_labels = list(enumerate(xdatum.discretized_labels))
        if n > MAXLABELS:
            ticks_labels = decimate_list(ticks_labels, MAXLABELS1)         # enforce max number of tick labels
        labels = [label for _, label in ticks_labels]
        rot = 90 if max([len(label) for label in xdatum.discretized_labels])*n > 60 else 0
        ax.set_xticks([x[0] for x in ticks_labels])
        ax.set_xticklabels(labels, rotation=rot)
        if len(ticks_labels) < n and n <= MAXTICKS:
            ax.set_xticks(range(n), minor=True)

    if ydatum.discretized_labels is not None:
        n = len(ydatum.discretized_labels)
        ticks_labels = list(enumerate(ydatum.discretized_labels))
        if n > MAXLABELS:
            ticks_labels = decimate_list(ticks_labels, MAXLABELS1)         # enforce max number of tick labels
        labels = [label for _, label in ticks_labels]
        ax.set_yticks([y[0] for y in ticks_labels])
        ax.set_yticklabels(labels)
        if len(ticks_labels) < n and n <= MAXTICKS:
            ax.set_yticks(range(n), minor=True)

    # colorbar?
    if color_minmax:
        import matplotlib.colors
        # discrete axis
        if caxis is not None and cdatum.is_discrete:
            norm = matplotlib.colors.Normalize(-0.5, len(color_key)-0.5)
            ticks = np.arange(len(color_key))
            colormap = matplotlib.colors.ListedColormap(color_key)
        # discretized axis
        else:
            norm = matplotlib.colors.Normalize(*color_minmax)
            colormap = matplotlib.colors.ListedColormap(color_key)
            # auto-mark colorbar, since it represents a continuous range of values
            ticks = None

        cb = fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=colormap), ax=ax, ticks=ticks)

        # adjust ticks for discrete axis
        if caxis is not None and cdatum.is_discrete:
            rot = 0
            # adjust fontsize for number of labels
            fs = max(options.fontsize*min(0.8, 20./len(color_labels)), 6)
            fontdict = dict(fontsize=fs)
            if max([len(lbl) for lbl in color_labels]) > 3 and len(color_labels) < 8:
                rot = 90
                fontdict['verticalalignment'] ='center'
            cb.ax.set_yticklabels(color_labels, rotation=rot, fontdict=fontdict)
        else:
            cb.ax.tick_params(labelsize=options.fontsize*0.8)

    fig.savefig(pngname, bbox_inches='tight')

    pylab.close()

    return pngname




#!/usr/bin/env python

import os
from setuptools import setup, find_packages

build_root = os.path.dirname(__file__)

requirements = [
"dask-ms[xarray]",
"dask[complete]",
"datashader>=0.12.0", 
"holoviews",
"matplotlib>2.2.3; python_version >= '3.5'",
"cmasher",
"future-fstrings",
"requests",
"MSUtils"
]

extras_require = {'testing': ['pytest', 'pytest-flake8']}

PACKAGE_NAME = 'shadems'
__version__ = '0.5.1.1'

def readme():
    """Get readme content for package long description"""
    with open(os.path.join(build_root, 'README.md')) as f:
        return f.read()

setup(name = PACKAGE_NAME,
    long_description=readme(),
    long_description_content_type="text/markdown",
    version = __version__,
    description = "Rapid Measurement Set plotting with dask-ms and datashader",
    author = "Ian Heywood & RATT",
    author_email = "ian.heywood@physics.ox.ac.uk",
    url = "https://github.com/ratt-ru/shadeMS",
    packages=find_packages(),
    install_requires = requirements,
    extras_require=extras_require,
    include_package_data = True,
    scripts=["bin/" + i for i in os.listdir("bin")],
    license="GNU GPL v2",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy"
    ]
)



from functools import reduce
from itertools import product
from operator import mul

import dask.array as da
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
import dask.dataframe as dd
import pandas as pd
import numpy as np


def start_ends(chunks):
    s = 0

    for c in chunks:
        e = s + c
        yield (s, e)
        s = e


def _create_dataframe(arrays, start, end, columns):
    index = None if start is None else np.arange(start, end)

    return pd.DataFrame({k: a.ravel() for k, a in zip(columns, arrays)},
                        index=index)

def dataframe_factory(out_ind, *arginds, columns=None):
    """
    Creates a dask Dataframe by broadcasting  *arginds
    against each other and then ravelling them.

    .. code-block:: python

        df = dataframe_factory(("row", "chan"),
                               x, ("row",),
                               y, ("chan",))

    Parameters
    ----------
    out_ind : sequence
        Output dimensions.
        e.g. :code:`(row, chan)`
    *arginds : Sequence of (:class:`dask.array.Array`, index)
        document me
    columns : sequence, optional
        Dataframe column names.
        Defaults to :code:`[x, y]`
    """
    if not len(arginds) % 2 == 0:
        raise ValueError("Must supply an index for each argument")

    args = arginds[::2]
    inds = arginds[1::2]

    if columns is None:
        columns = ['x', 'y'] + ["c%d" % i for i in range(len(args) - 2)]
    else:
        if (not isinstance(columns, (tuple, list)) and
            len(columns) != len(args)):

            raise ValueError("Columns must be a tuple/list of columns "
                             "matching the number of arrays")

    have_nan_chunks = False

    new_args = []

    for a, (arg, ind) in enumerate(zip(args, inds)):
        if not all(i in out_ind for i in ind):
            raise ValueError("Argument %d dimensions not in out_ind" % a)

        if not len(ind) == arg.ndim:
            raise ValueError("Argument %d len(ind) != arg.ndim" % a)

        have_nan_chunks = (any(np.isnan(c) for dc in arg.chunks for c in dc) or
                           have_nan_chunks)

        # Generate slicing tuple that will expand arg up to full resolution
        expand = tuple(slice(None) if i in ind else None for i in out_ind)
        new_args.append(arg[expand])

    # Create meta data so that blockwise doesn't call
    # np.broadcast_arrays and fall over on the tuple
    # of arrays that it returns
    dtype = np.result_type(*args)
    meta = np.empty((0,) * len(out_ind), dtype=dtype)

    blockargs = (v for pair in ((a, out_ind) for a in new_args) for v in pair)

    bcast = da.blockwise(np.broadcast_arrays, out_ind,
                         *blockargs,
                         subok=True,
                         align_arrays=not have_nan_chunks,
                         meta=meta,
                         dtype=dtype)

    # Now create a dataframe from the broadcasted arrays
    # with lower-level dask graph API

    # Flattened list of broadcast array keys
    # We'll use this to generate a 1D (ravelled) dataframe
    keys = product((bcast.name,), *(range(b) for b in bcast.numblocks))
    name = "dataframe-" + tokenize(bcast)

    # dictionary defining the graph for this part of the operation
    layers = {}

    if have_nan_chunks:
        # We can't create proper indices if we don't known our chunk sizes
        divisions = [None]

        for i, key in enumerate(keys):
            layers[(name, i)] = (_create_dataframe, key, None, None, columns)
            divisions.append(None)
    else:
        # We do know all our chunk sizes, create reasonable dataframe indices
        start_idx = 0
        divisions = [0]

        expr = ((e - s for s, e in start_ends(dim_chunks))
                for dim_chunks in bcast.chunks)
        chunk_sizes = (reduce(mul, shape, 1) for shape in product(*expr))
        chunk_ranges = start_ends(chunk_sizes)

        for i, (key, (start, end)) in enumerate(zip(keys, chunk_ranges)):
            layers[(name, i)] = (_create_dataframe, key, start, end, columns)
            start_idx += end - start
            divisions.append(start_idx)

    assert len(layers) == bcast.npartitions
    assert len(divisions) == bcast.npartitions + 1

    # Create the HighLevelGraph
    graph = HighLevelGraph.from_collections(name, layers, [bcast])
    # Metadata representing the broadcasted and ravelled data

    meta = pd.DataFrame(data={k: np.empty((0,), dtype=a.dtype)
                              for k, a in zip(columns, args)},
                        columns=columns)

    # Create the actual Dataframe
    return dd.DataFrame(graph, name, meta=meta, divisions=divisions)


def multicol_dataframe_factory(out_ind, arrays, array_dims):
    """
    Creates a dask Dataframe by broadcasting arrays (given by the arrays dict-like object)
    against each other and then ravelling them. The array_indices mapping specifies which indices
    the arrays have

    .. code-block:: python

        df = dataframe_factory(("row", "chan"), {'x': x, 'y': y}, {x: ("row",), y: ("chan",)})

    Parameters
    ----------
    out_ind : sequence
        Output dimensions.
        e.g. :code:`(row, chan)`
    """
    columns = list(arrays.keys())

    have_nan_chunks = None
    expand = {}
    barr = {}
    # build up list of arguments for blockwise call below
    blockwise_args = [np.broadcast_arrays, out_ind]

    for col, arr in arrays.items():
        if col not in array_dims:
            raise ValueError(f"{col} dimensions not specified")
        arr_ind = array_dims[col]
        if not all(i in out_ind for i in arr_ind):
            raise ValueError(f"{col} dimensions not in out_ind")
        if not len(arr_ind) == arr.ndim:
            raise ValueError(f"len({col}_ind) != {col}.ndim")
        have_nan_chunks = have_nan_chunks or any(np.isnan(c) for dc in arr.chunks for c in dc)

        # Generate slicing tuples that will expand arr up to the full
        # resolution
        expand[col] = tuple(slice(None) if i in arr_ind else None for i in out_ind)
        # broadcast vesion of array
        barr[col] = arr[expand[col]]

        blockwise_args += [barr[col], out_ind]

    # Create meta data so that blockwise doesn't call
    # np.broadcast_arrays and fall over on the tuple
    # of arrays that it returns
    dtype = np.result_type(*arrays.values())
    meta = np.empty((0,) * len(out_ind), dtype=dtype)

    bcast = da.blockwise(*blockwise_args,
                         align_arrays=not have_nan_chunks,
                         meta=meta,
                         dtype=dtype)

    # Now create a dataframe from the broadcasted arrays
    # with lower-level dask graph API

    # Flattened list of broadcast array keys
    # We'll use this to generate a 1D (ravelled) dataframe
    keys = product((bcast.name,), *(range(b) for b in bcast.numblocks))
    name = "dataframe-" + tokenize(bcast)

    # dictionary defining the graph for this part of the operation
    layers = {}

    if have_nan_chunks:
        # We can't create proper indices if we don't known our chunk sizes
        divisions = [None]

        for i, key in enumerate(keys):
            layers[(name, i)] = (_create_dataframe, key, None, None)
            divisions.append(None)
    else:
        # We do know all our chunk sizes, create reasonable dataframe indices
        start_idx = 0
        divisions = [0]

        expr = ((e - s for s, e in start_ends(dim_chunks))
                for dim_chunks in bcast.chunks)
        chunk_sizes = (reduce(mul, shape, 1) for shape in product(*expr))
        chunk_ranges = start_ends(chunk_sizes)

        for i, (key, (start, end)) in enumerate(zip(keys, chunk_ranges)):
            layers[(name, i)] = (_create_dataframe, key, start, end)
            start_idx += end - start
            divisions.append(start_idx)

    assert len(layers) == bcast.npartitions
    assert len(divisions) == bcast.npartitions + 1

    # Create the HighLevelGraph
    graph = HighLevelGraph.from_collections(name, layers, [bcast])
    # Metadata representing the broadcasted and ravelled data
    meta = pd.DataFrame(data={col: np.empty((0,), dtype=arr.dtype) for col, arr in arrays.items()},
                        columns=columns)

    # Create the actual Dataframe
    return dd.DataFrame(graph, name, meta=meta, divisions=divisions)



import logging
import os

# set default number of renderers to half the available cores
DEFAULT_CNUM = 16
DEFAULT_NUM_RENDERS = max(1, len(os.sched_getaffinity(0))//2)

from .__main__ import cli, parse_plot_spec, parse_slice_spec  # noqa

# create logger with 'spam_application'
log = logging.getLogger('shadems')
log.setLevel(logging.DEBUG)
# create file handler which logs even debug messages
log_file_handler = logging.FileHandler('log-shadems.txt')
log_file_handler.setLevel(logging.DEBUG)
# create console handler with a higher log level
log_console_handler = logging.StreamHandler()
log_console_handler.setLevel(logging.INFO)
# create formatter and add it to the handlers
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                              "%Y-%m-%d %H:%M:%S")
log_file_handler.setFormatter(formatter)
log_console_handler.setFormatter(formatter)
# add the handlers to the logger
log.addHandler(log_file_handler)
log.addHandler(log_console_handler)


def separator():
    log.info('------------------------------------------------------')
